from typing import Any, Optional, Iterator
from collections import UserDict
from omnibelt import filter_duplicates

from .abstract import AbstractGang, AbstractGig
from .errors import GadgetFailure, ApplicationAmbiguityError
from .gigs import GigBase


class GangBase(GigBase, AbstractGang):
	"""
	The GangBase class is a base for custom gangs to enable remapping the gizmo names.

	Attributes:
		_current_context (Optional[AbstractGig]): The current context associated with this gang.
	"""

	_current_context: Optional[AbstractGig]

	def __init__(self, *, apply: Optional[dict[str, str]] = None, **kwargs):
		"""
		Initializes a new instance of the GangBase class.

		Args:
			apply (Optional[dict[str, str]]): gizmo relabeling. If not provided, an empty dictionary will be used.
			kwargs: unused, passed to super.
		"""
		if apply is None:
			apply = {}
		super().__init__(**kwargs)
		self._raw_apply = apply
		self._raw_reverse_apply = None
		self._current_context = None

	def _gizmos(self) -> Iterator[str]:
		"""
		Lists gizmos produced by self using internal names.

		Returns:
			Iterator[str]: An iterator over the gizmos.
		"""
		yield from super().gizmos()

	def gizmos(self) -> Iterator[str]:
		"""
		Lists gizmos produced by self using external names.

		Returns:
			Iterator[str]: An iterator over the gizmos.
		"""
		for gizmo in self._gizmos():
			yield self.gizmo_to(gizmo)

	@staticmethod
	def _infer_external2internal(raw: dict[str, str], products: Iterator[str]) -> dict[str, str]:
		"""
		Infers the external to internal gizmo mapping (reverse) from the provided raw mapping and products.

		Note, that this method generally shouldn't have to be called by subclasses.

		Args:
			raw (dict[str, str]): The gizmo mapping of internal to external names.
			products (Iterator[str]): An iterator over the products (expected to be `_gizmos`)

		Returns:
			dict[str, str]: The corresponding external to internal gizmo mapping.
		"""
		reverse = {}
		for product in products:
			if product in raw:
				external = raw[product]
				if external in reverse:
					raise ApplicationAmbiguityError(product, [reverse[external], product])
				reverse[external] = product
		return reverse

	def _update_external2internal(self, raw: dict[str, str]):
		self._raw_apply = raw
		self._raw_reverse_apply = None

	def gizmo_from(self, gizmo: str) -> str:
		"""
		Converts an external gizmo to its internal names.

		Used primarily when grabbing a gizmo using self as the context.

		Args:
			gizmo (str): The external gizmo name.

		Returns:
			str: The internal name of the gizmo.
		"""
		if self._raw_reverse_apply is None:
			self._raw_reverse_apply = self._infer_external2internal(self._raw_apply, self._gizmos())
		return self._raw_reverse_apply.get(gizmo, gizmo)

	def gizmo_to(self, gizmo: str) -> str:
		"""
		Converts an internal gizmo to its external representation.

		Used primarily to grab a gizmo from an external context if self fails.

		Args:
			gizmo (str): The internal gizmo name.

		Returns:
			str: The external name of the gizmo.
		"""
		return self._raw_apply.get(gizmo, gizmo)

	def _grab_from_fallback(self, error: GadgetFailure, ctx: Optional[AbstractGig], gizmo: str) -> Any:
		"""
		Handles a GadgetFailure when trying to grab a gizmo from the context.

		Args:
			error (GadgetFailure): The GadgetFailure that occurred.
			ctx (Optional[AbstractGig]): The context from which to grab the gizmo.
			gizmo (str): The name of the gizmo to grab.

		Returns:
			Any: The result of the fallback operation.
		"""
		return super()._grab_from_fallback(error, self._current_context, self.gizmo_to(gizmo))

	def grab_from(self, ctx: Optional[AbstractGig], gizmo: str) -> Any:
		"""
		Tries to grab a gizmo from the context.

		Args:
			ctx (Optional[AbstractGig]): The context from which to grab the gizmo.
			gizmo (str): The name of the gizmo to grab.

		Returns:
			Any: The grabbed gizmo.
		"""
		if ctx is not None and ctx is not self and self._current_context is None:
			assert self._current_context is None, f'Context already set to {self._current_context}'
			self._current_context = ctx
		if ctx is self._current_context:
			gizmo = self.gizmo_from(gizmo)  # convert to internal gizmo
		return super().grab_from(self, gizmo)





from typing import Any, Optional, Iterator
from collections import UserDict
from omnibelt import filter_duplicates

from .abstract import AbstractGadget, AbstractGaggle, AbstractGig, AbstractGang, AbstractGadgetError
from .errors import GadgetFailure, MissingGadget, AssemblyError, GrabError
from .gadgets import GadgetBase


class GigBase(GadgetBase, AbstractGig):
	"""
	The GigBase class is a subclass of GadgetBase and AbstractGig. It provides methods to handle gizmo grabbing and packaging.

	Attributes:
		_GrabError (Exception): The exception to be raised when a grab operation fails.
	"""

	_GrabError = GrabError

	def _grab_from_fallback(self, error: Exception, ctx: Optional[AbstractGig], gizmo: str) -> Any:
		"""
		Handles a GadgetFailure when trying to grab a gizmo from the context.

		Args:
			error (Exception): The exception that occurred during the grab operation.
			ctx (Optional[AbstractGig]): The context from which to grab the gizmo.
			gizmo (str): The name of the gizmo to grab.

		Returns:
			Any: The result of the fallback operation.

		Raises:
			_GrabError: If the error is a GrabError or if the context is None or self.
			error: If the error is not a GrabError.
		"""
		if isinstance(error, AbstractGadgetError):
			if isinstance(error, GrabError) or ctx is None or ctx is self:
				raise self._GrabError(gizmo, error) from error
			else:
				return ctx.grab(gizmo)
		raise error from error

	def _grab(self, gizmo: str) -> Any:
		"""
		Grabs a gizmo from self.

		Args:
			gizmo (str): The name of the gizmo to grab.

		Returns:
			Any: The grabbed gizmo.
		"""
		return super().grab_from(self, gizmo)

	def grab_from(self, ctx: Optional[AbstractGig], gizmo: str) -> Any:
		"""
		Tries to grab a gizmo from the context.

		Args:
			ctx (Optional[AbstractGig]): The context from which to grab the gizmo.
			gizmo (str): The name of the gizmo to grab.

		Returns:
			Any: The grabbed gizmo.
		"""
		try:
			out = self._grab(gizmo)
		except Exception as error:
			out = self._grab_from_fallback(error, ctx, gizmo)
		return self.package(out, gizmo=gizmo)

	def package(self, val: Any, *, gizmo: Optional[str] = None) -> Any:
		"""
		Packages a value with an optional gizmo.

		Args:
			val (Any): The value to be packaged.
			gizmo (Optional[str]): The name of the gizmo. Defaults to None.

		Returns:
			Any: The packaged value.
		"""
		return val


class CacheGig(GigBase, UserDict):
	"""
	The CacheGig class is a subclass of GigBase and UserDict. It provides methods to handle gizmo caching.

	Attributes:
		_gizmo_type (Optional[type]): The type of the gizmo. Defaults to None.
	"""

	_gizmo_type = None

	def __setitem__(self, key, value):
		"""
		Sets an item in the dictionary.

		Args:
			key: The key of the item.
			value: The value of the item.
		"""
		if self._gizmo_type is not None:
			key = self._gizmo_type(key)
		super().__setitem__(key, value)

	def __repr__(self):
		"""
		Returns a string representation of the CacheGig instance.

		Returns:
			str: A string representation of the CacheGig instance.
		"""
		gizmos = [(gizmo if self.is_cached(gizmo) else '{' + str(gizmo) + '}') for gizmo in self.gizmos()]
		return f'{self.__class__.__name__}({", ".join(gizmos)})'

	def gizmos(self) -> Iterator[str]:
		"""
		Lists gizmos produced by self.

		Returns:
			Iterator[str]: An iterator over the gizmos.
		"""
		yield from filter_duplicates(self.cached(), super().gizmos())

	def is_cached(self, gizmo: str) -> bool:
		"""
		Checks if a gizmo is cached.

		Args:
			gizmo (str): The name of the gizmo to check.

		Returns:
			bool: True if the gizmo is cached, False otherwise.
		"""
		return gizmo in self.data

	def cached(self) -> Iterator[str]:
		"""
		Lists the cached gizmos.

		Returns:
			Iterator[str]: An iterator over the cached gizmos.
		"""
		yield from self.data.keys()

	def clear_cache(self) -> None:
		"""
		Clears the cache.
		"""
		self.data.clear()

	def grab_from(self, ctx: Optional[AbstractGig], gizmo: str) -> Any:
		"""
		Tries to grab a gizmo from the context.

		Args:
			ctx (Optional[AbstractGig]): The context from which to grab the gizmo.
			gizmo (str): The name of the gizmo to grab.

		Returns:
			Any: The grabbed gizmo.
		"""
		if gizmo in self.data:
			return self.data[gizmo]
		val = super().grab_from(ctx, gizmo)
		self.data[gizmo] = val  # cache packaged val
		return val

class GroupCache(CacheGig):
	"""
	The GroupCache class is a subclass of CacheGig. It provides methods to handle gizmo caching with group support.

	Attributes:
		_group_cache (dict): A dictionary to store group caches.
	"""

	def __init__(self, *args, group_cache=None, **kwargs):
		"""
		Initializes a new instance of the GroupCache class.

		Args:
			args: Variable length argument list.
			group_cache (Optional[dict]): A dictionary of group caches. If not provided, an empty dictionary will be used.
			kwargs: Arbitrary keyword arguments.
		"""
		if group_cache is None:
			group_cache = {}
		super().__init__(*args, **kwargs)
		self._group_cache = group_cache

	def is_cached(self, gizmo: str) -> bool:
		"""
		Checks if a gizmo is cached in either the main cache or any of the group caches.

		Args:
			gizmo (str): The name of the gizmo to check.

		Returns:
			bool: True if the gizmo is cached, False otherwise.
		"""
		if super().is_cached(gizmo):
			return True
		for group, cache in self._group_cache.items():
			for key in cache:
				if group.gizmo_to(key) == gizmo:
					return True
		return False

	def cached(self) -> Iterator[str]:
		"""
		Lists the cached gizmos in both the main cache and the group caches.

		Returns:
			Iterator[str]: An iterator over the cached gizmos.
		"""
		def _group_cached():
			for group, cache in self._group_cache.items():
				for key in cache:
					yield group.gizmo_to(key)
		yield from filter_duplicates(super().cached(), _group_cached())

	def check_group_cache(self, group: AbstractGang, gizmo: str):
		"""
		Checks a group cache for a gizmo.

		Args:
			group (AbstractGroup): The group to check.
			gizmo (str): The name of the gizmo to check.

		Returns:
			Any: The cached value of the gizmo in the group cache.
		"""
		return self._group_cache[group][gizmo]

	def update_group_cache(self, group: AbstractGang, gizmo: str, val: Any):
		"""
		Updates a group cache with a gizmo and its value.

		Args:
			group (AbstractGroup): The group to update.
			gizmo (str): The name of the gizmo to update.
			val (Any): The value of the gizmo to update.
		"""
		if self._gizmo_type is not None:
			gizmo = self._gizmo_type(gizmo)
		self._group_cache.setdefault(group, {})[gizmo] = val

	def clear_cache(self, *, clear_group_caches=True) -> None:
		"""
		Clears the cache and optionally the group caches.

		Args:
			clear_group_caches (bool): Whether to clear the group caches. Defaults to True.
		"""
		super().clear_cache()
		if clear_group_caches:
			self._group_cache.clear()




from typing import Tuple, List, Dict, Optional, Union, Any, Callable, Sequence, Iterator, Iterable, Type, Set
import math

from ..features import Prepared
from ..tools.moguls import BatchMogul, EpochStatMogul, SeedingMogul, SimpleMogul
from ..parameters import hparam, Parameterized

from .abstract import AbstractProgression, AbstractBatchable, AbstractBatch
from .errors import BudgetExceeded
from .progression import ProgressionBase



_inf = float('inf')


class BudgetLoader(Parameterized, SeedingMogul, SimpleMogul, AbstractProgression):
	batch_size = hparam(required=True)

	sample_limit = hparam(None)
	batch_limit = hparam(None)
	@hparam
	def epoch_limit(self):
		if self.epoch_size is not None and self.sample_limit is None and self.batch_limit is None:
			if self.strict_limit is None:
				self.strict_limit = True
			return 1


	strict_limit = hparam(None, hidden=True)
	strict_batch_size = hparam(None, hidden=True)

	shuffle = hparam(False, hidden=True)
	@hparam
	def infinite(self): # TODO: deal with circular hparam dependencies
		return self.sample_limit is None and self.batch_limit is None and self.epoch_limit is None

	# use_pbar = hparam(False, inherit=True, hidden=True) # TODO
	# pbar_samples = hparam(True, inherit=True, hidden=True)


	@hparam
	def epoch_size(self):
		src = self.source
		if src is not None:
			try:
				return src.size
			except AttributeError:
				pass


	def __init__(self, source: AbstractBatchable = None, **kwargs):
		super().__init__(**kwargs)
		self._loader = None
		self._full_epochs = None
		self._total_samples = None
		self._total_batches = None
		if source is not None:
			self.set_source(source)


	def set_source(self, source: AbstractBatchable, **kwargs):
		if isinstance(source, AbstractProgression):
			source = source.source
		self._loader = source.iterate(self.batch_size, **kwargs)
		self._is_ready = False
		self.prepare()
		return self


	@property
	def source(self) -> AbstractBatchable:
		return self._loader.source


	# def sources(self) -> Iterator['AbstractBatchable']:
	# 	yield self.source


	@property
	def current_batch(self) -> AbstractBatch:
		return self._loader.current_batch


	_BudgetExceeded = BudgetExceeded

	def create_batch(self, size: Optional[int] = None) -> AbstractBatch:
		if self.done():
			raise self._BudgetExceeded
		if size is None:
			size = self.batch_size
		if self.remaining_samples is not None:
			size = min(size, self.remaining_samples)
		ctx = self._loader.create_batch(size)
		ctx.set_progress(self)
		return ctx


	def _prepare(self, *args, **kwargs):
		super()._prepare(*args, **kwargs)

		if not self.infinite:
			if self.epoch_size is None:
				self._total_samples, self._total_batches = self.compute_budget(
					samples_per_batch=self.batch_size, strict_batch_size=self.strict_batch_size,
					sample_limit=self.sample_limit, batch_limit=self.batch_limit, strict_limit=self.strict_limit,
				)
			else:
				self._total_samples, self._total_batches, self._full_epochs = self.compute_epoch_budget(
					dataset_size=self.epoch_size, samples_per_batch=self.batch_size,
					strict_batch_size=self.strict_batch_size,
					epochs=self.epoch_limit, sample_limit=self.sample_limit,
					batch_limit=self.batch_limit, strict_limit=self.strict_limit
				)


	@staticmethod
	def compute_budget(samples_per_batch, strict_batch_size=False,
	                   sample_limit=None, batch_limit=None, strict_limit=True):
		if sample_limit is None and batch_limit is None:
			return None, None  # infinite

		total_samples = None
		if batch_limit is not None:
			total_samples = batch_limit * samples_per_batch
		if sample_limit is not None:

			remainder = sample_limit % samples_per_batch
			total = sample_limit - remainder
			if remainder > 0:
				if strict_limit and not strict_batch_size:
					total += remainder
				elif not strict_limit:
					total += samples_per_batch

			# total = sample_limit - (sample_limit % samples_per_batch)
			# remainder = sample_limit % samples_per_batch
			# if remainder > 0:
			# 	if strict_limit and not strict_batch_size:
			# 		total += remainder
			# 	elif not strict_limit:
			# 		total += samples_per_batch
			if total_samples is None or total < total_samples:
				total_samples = total

		total_batches = total_samples // samples_per_batch
		remainder = total_samples % samples_per_batch
		if not strict_batch_size and remainder > 0:
			total_batches += 1

		return total_samples, total_batches


	@staticmethod
	def compute_epoch_budget(dataset_size, samples_per_batch, strict_batch_size=True,
	                   epochs=None, sample_limit=None, batch_limit=None, strict_limit=True):
		if epochs is None and sample_limit is None and batch_limit is None:
			return None, None, None  # infinite

		strict_limit = bool(strict_limit)
		strict_batch_size = bool(strict_batch_size)

		samples_per_epoch = dataset_size - int(strict_batch_size) * (dataset_size % samples_per_batch)
		batches_per_epoch = int(math.ceil(samples_per_epoch / samples_per_batch))

		total_samples = None if epochs is None else samples_per_epoch * epochs
		if batch_limit is not None:
			total = (batch_limit % batches_per_epoch) * samples_per_batch \
			        + (batch_limit // batches_per_epoch) * samples_per_epoch
			if total_samples is None or total < total_samples:
				total_samples = total
		if sample_limit is not None:
			remainder = sample_limit % samples_per_batch
			total = sample_limit - remainder
			if remainder > 0:
				if strict_limit and not strict_batch_size:
					total += remainder
				elif not strict_limit:
					total += samples_per_batch
			# total = samples_per_epoch * (sample_limit // samples_per_epoch)
			# remainder = sample_limit % samples_per_epoch
			# total += samples_per_batch * (remainder // samples_per_batch)
			# remainder = remainder % samples_per_batch
			# if remainder > 0:
			# 	if strict_limit and not strict_batch_size:
			# 		total += remainder
			# 	elif not strict_limit:
			# 		total += samples_per_batch
			if total_samples is None or total < total_samples:
				total_samples = total

		full_epochs = total_samples // samples_per_epoch
		remainder = total_samples % samples_per_epoch
		total_batches = full_epochs * batches_per_epoch + remainder // samples_per_batch
		remainder = remainder % samples_per_batch
		if not strict_batch_size and remainder > 0:
			total_batches += 1

		return total_samples, total_batches, full_epochs


	@property
	def sample_count(self) -> int:
		return self._loader.sample_count
	@property
	def batch_count(self) -> int:
		return self._loader.batch_count


	@property
	def budget_samples(self) -> Optional[int]:
		return self._total_samples
	@property
	def budget_batches(self) -> Optional[int]:
		return self._total_batches


	@property
	def full_epochs(self) -> Optional[int]:
		return self._full_epochs
	@property
	def completed_epochs(self) -> int:
		if self.epoch_size is not None:
			return self.sample_count // self.epoch_size


	@property
	def current_epoch(self) -> int:
		if self.epoch_size is not None:
			return self.completed_epochs + 1


	@property
	def remaining_samples(self):
		if self._total_samples is None:
			return _inf
		return self._total_samples - self.sample_count
	@property
	def remaining_batches(self):
		if self._total_batches is None:
			return _inf
		return self._total_batches - self.batch_count
	@property
	def remaining_epochs(self): # only lists full epochs
		if self.epoch_size is not None:
			if self._full_epochs is None:
				return _inf
			return self._full_epochs - self.completed_epochs
	@property
	def remaining_samples_in_epoch(self):
		if self.epoch_size is not None:
			return self.epoch_size - self.sample_count % self.epoch_size
	@property
	def remaining_batches_in_epoch(self): # only lists full batches
		if self.epoch_size is not None:
			return self.remaining_samples_in_epoch // self.batch_size


	def done(self) -> bool:
		todo = self.remaining_samples
		return todo is None or todo <= 0























#
# _inf = float('inf')
#
# class BudgetProgressionBase(AbstractProgression):
# 	def __init__(self, *, sample_limit=None, batch_limit=None, strict_limit=True, strict_batch_size=False, **kwargs):
# 		super().__init__(**kwargs)
# 		self._sample_limit = sample_limit
# 		self._batch_limit = batch_limit
#
# 		self._strict_limit = strict_limit
# 		self._strict_batch_size = strict_batch_size
#
# 		self._total_samples = None
# 		self._total_batches = None
#
#
# 	def _prepare(self, *args, **kwargs):
# 		super()._prepare(*args, **kwargs)
# 		self._total_samples, self._total_batches = self.compute_budget(
# 			samples_per_batch=self.batch_size, strict_batch_size=self._strict_batch_size,
# 			sample_limit=self._sample_limit, batch_limit=self._batch_limit, strict_limit=self._strict_limit,
# 		)
#
#
# 	@staticmethod
# 	def compute_budget(samples_per_batch, strict_batch_size=False,
# 	                   sample_limit=None, batch_limit=None, strict_limit=True):
# 		if sample_limit is None and batch_limit is None:
# 			return None, None  # infinite
#
# 		total_samples = None
# 		if batch_limit is not None:
# 			total_samples = batch_limit * samples_per_batch
# 		if sample_limit is not None:
# 			total = sample_limit - (sample_limit % samples_per_batch)
# 			remainder = sample_limit % samples_per_batch
# 			if remainder > 0:
# 				if strict_limit and not strict_batch_size:
# 					total += remainder
# 				elif not strict_limit:
# 					total += samples_per_batch
# 			if total_samples is None or total < total_samples:
# 				total_samples = total
#
# 		total_batches = total_samples // samples_per_batch
# 		remainder = total_samples % samples_per_batch
# 		if not strict_batch_size and remainder > 0:
# 			total_batches += 1
#
# 		return total_samples, total_batches
#
#
# 	@property
# 	def budget_samples(self) -> Optional[int]:
# 		return self._total_samples
# 	@property
# 	def budget_batches(self) -> Optional[int]:
# 		return self._total_batches
#
#
# 	@property
# 	def remaining_samples(self):
# 		if self._total_samples is None:
# 			return _inf
# 		return self._total_samples - self.sample_count
# 	@property
# 	def remaining_batches(self):
# 		if self._total_batches is None:
# 			return _inf
# 		return self._total_batches - self.batch_count
#
#
#
# class EpochBudgetProgressionBase(BudgetProgressionBase, EpochStatMogul):
# 	def __init__(self, *, epoch_limit=None, **kwargs):
# 		super().__init__(**kwargs)
# 		self._epoch_limit = epoch_limit
# 		self._full_epochs = None
#
#
# 	def _prepare(self, *args, **kwargs):
# 		super()._prepare(*args, **kwargs)
#
# 		if self._epoch_limit is None \
# 				and self._sample_limit is None \
# 				and self._batch_limit is None:
# 			self._epochs = 1
#
# 		self._total_samples, self._total_batches, self._full_epochs = self.compute_epoch_budget(
# 			dataset_size=self.epoch_size, samples_per_batch=self.batch_size,
# 			strict_batch_size=self._strict_batch_size,
# 			epochs=self._epochs, sample_limit=self._sample_limit,
# 			batch_limit=self._batch_limit, strict_limit=self._strict_limit
# 		)
#
#
# 	@staticmethod
# 	def compute_epoch_budget(dataset_size, samples_per_batch, strict_batch_size=True,
# 	                   epochs=None, sample_limit=None, batch_limit=None, strict_limit=True):
# 		if epochs is None and sample_limit is None and batch_limit is None:
# 			return None, None, None  # infinite
#
# 		samples_per_epoch = dataset_size - int(strict_batch_size) * (dataset_size % samples_per_batch)
# 		batches_per_epoch = int(math.ceil(samples_per_epoch / samples_per_batch))
#
# 		total_samples = None if epochs is None else samples_per_epoch * epochs
# 		if batch_limit is not None:
# 			total = (batch_limit % batches_per_epoch) * samples_per_batch \
# 			        + (batch_limit // batches_per_epoch) * samples_per_epoch
# 			if total_samples is None or total < total_samples:
# 				total_samples = total
# 		if sample_limit is not None:
# 			total = samples_per_epoch * (sample_limit // samples_per_epoch)
# 			remainder = sample_limit % samples_per_epoch
# 			total += samples_per_batch * (remainder // samples_per_batch)
# 			remainder = remainder % samples_per_batch
# 			if remainder > 0:
# 				if strict_limit and not strict_batch_size:
# 					total += remainder
# 				elif not strict_limit:
# 					total += samples_per_batch
# 			if total_samples is None or total < total_samples:
# 				total_samples = total
#
# 		full_epochs = total_samples // samples_per_epoch
# 		remainder = total_samples % samples_per_epoch
# 		total_batches = full_epochs * batches_per_epoch + remainder // samples_per_batch
# 		remainder = remainder % samples_per_batch
# 		if not strict_batch_size and remainder > 0:
# 			total_batches += 1
#
# 		return total_samples, total_batches, full_epochs
#
#
# 	@property
# 	def full_epochs(self) -> Optional[int]:
# 		return self._full_epochs
#
#
# 	@property
# 	def remaining_epochs(self):
# 		if self._full_epochs is None:
# 			return float('inf')
# 		return self._full_epochs - self.current_epoch





# class BarredProgression(ProgressionBase, ProgressBarred):
# 	def __init__(self, source, use_pbar=False, pbar_samples=True, **kwargs):
# 		super().__init__(source=source, **kwargs)
# 		self._use_pbar = use_pbar
# 		self._pbar_samples = pbar_samples
#
# 	def _create_pbar(self, *, unit=None, **kwargs):
# 		if unit is None:
# 			unit = 'smpl' if self._pbar_samples else 'batch'
# 		return super()._create_pbar(unit=unit, **kwargs)
#
# 	def _prepare(self, *args, **kwargs):
# 		super()._prepare(*args, **kwargs)
# 		if self._use_pbar:
# 			self._create_pbar()
#
# 	def _next_batch(self):
# 		pbar = self._pbar
# 		try:
# 			batch = super()._next_batch()
# 		except StopIteration:
# 			if pbar is not None:
# 				pbar.close()
# 			raise
# 		else:
# 			if pbar is not None:
# 				pbar.update(self.batch_size if self._pbar_samples else 1)
# 			return batch
#
# 	def set_description(self, desc):
# 		if self._pbar is not None:
# 			self._pbar.set_description(desc)
#
#
#
# class TrackedProgression(BarredProgression, AbstractBudgetProgression): # TODO: add a progress bar
# 	def _create_pbar(self, total=None, **kwargs):
# 		if total is None:
# 			total = self.budget_samples if self._pbar_samples else self.budget_batches
# 		return super()._create_pbar(total=total, **kwargs)




























from typing import Tuple, List, Dict, Optional, Union, Any, Callable, Sequence, Iterator, Iterable

from omnibelt import unspecified_argument, duplicate_instance, get_printer, primitives


from ..features import Prepared, Seedable
from ..persistent import AbstractFingerprinted
from ..tools.abstract import AbstractTool, AbstractKit, AbstractSourcedKit, AbstractDynamicKit, \
	AbstractSpaced, AbstractContext, AbstractMogul, AbstractScope, AbstractSchema
from ..tools.errors import MissingGizmoError
from ..tools.moguls import BatchMogul, IteratorMogul, CreativeMogul

from .errors import MissingBuffer


prt = get_printer(__file__)


class AbstractData(Prepared, AbstractFingerprinted): # TODO: make fingerprinted
	def copy(self):
		return duplicate_instance(self) # shallow copy


	def _title(self):
		return getattr(self, 'name', self.__class__.__name__)
	@property
	def title(self):
		return self._title()


	def __str__(self):
		return self.title


	def __repr__(self):
		return self.title



class AbstractCountableData(AbstractData):
	def _title(self):
		return f'{super()._title()}[{self.size}]'


	@property
	def size(self):
		raise NotImplementedError



class AbstractDataSource(AbstractTool, AbstractSpaced, AbstractData):
	# @classmethod
	# def _parse_context(cls, context: AbstractContext):
	# 	return context
	pass



class AbstractDataRouter(AbstractDynamicKit, AbstractDataSource):
	_MissingBuffer = MissingBuffer


	def include(self, *sources: AbstractTool, **buffers: AbstractDataSource) -> 'AbstractDynamicKit':
		raise NotImplementedError


	def _prepare(self, source=None, **kwargs):
		super()._prepare(source=source, **kwargs)
		for buffer in self.buffers():
			if isinstance(buffer, Prepared):
				buffer.prepare()


	def named_buffers(self) -> Iterator[Tuple[str, 'AbstractDataSource']]:
		raise NotImplementedError


	def buffers(self) -> Iterator['AbstractDataSource']:
		for name, buffer in self.named_buffers():
			yield buffer


	def buffer_names(self) -> Iterator[str]:
		for name, buffer in self.named_buffers():
			yield name


	def tools(self) -> Iterator['AbstractTool']:
		yield from self.buffers()
		yield from super().tools()


	def get_buffer(self, gizmo: str, default: Optional[Any] = unspecified_argument):
		raise NotImplementedError


	# def register_buffer(self, gizmo: str, buffer: Optional[AbstractTool] = None):
	# 	if not isinstance(gizmo, primitives):
	# 		raise TypeError(f'Gizmo must be a primitive type, not {type(gizmo)}')
	# 	return self.include(**{gizmo:buffer})


	# def _register_multi_buffer(self, buffer: AbstractTool, *gizmos: str): # TODO: move downstream (-> mixin)
	# 	for gizmo in gizmos:
	# 		self.register_buffer(gizmo, buffer)


	def remove_buffer(self, name):
		raise NotImplementedError


	def __str__(self):
		return f'{super().__str__()}({", ".join(map(str, self.gizmos()))})'



class AbstractView(AbstractContext, AbstractDataSource):
	def __init__(self, source: AbstractDataRouter = None, **kwargs):
		super().__init__(**kwargs)


	def _prepare(self, **kwargs):
		super()._prepare(**kwargs)
		self.source.prepare()


	def _title(self):
		return f'{super()._title()}{"<" + self.source.title + ">" if self.source is not None else ""}'


	@property
	def source(self):
		raise NotImplementedError


	def get_from(self, ctx: AbstractContext, gizmo: str):
		return self.source.get_from(ctx, gizmo)



class AbstractRouterView(AbstractView, AbstractDataRouter):
	def named_buffers(self) -> Iterator[Tuple[str, 'AbstractDataSource']]:
		yield from self.source.named_buffers()


	def buffers(self) -> Iterator['AbstractDataSource']:
		yield from self.source.buffers()


	# def gizmos(self) -> Iterator[str]:
	# 	yield from self.source.gizmos()
	#
	#
	# def tools(self) -> Iterator['AbstractTool']:
	# 	yield from self.source.tools()
	#
	#
	# def vendors(self, gizmo: str) -> Iterator['AbstractTool']:
	# 	yield from self.source.vendors(gizmo)
	#
	#
	# def has_gizmo(self, gizmo):
	# 	return self.source.has_gizmo(gizmo)


	def get_buffer(self, gizmo: str, default: Optional[Any] = unspecified_argument):
		return self.source.get_buffer(gizmo, default=default)


	# def validate_context(self, selection):
	# 	return self.source.validate_context(selection)




####################



class AbstractViewable(AbstractDataRouter):
	_View = None
	def view(self, **kwargs):
		return self._View(self, **kwargs)



class AbstractViewableRouterView(AbstractRouterView, AbstractViewable):
	def view(self, **kwargs):
		if self._View is None:
			return self.source._View(self, **kwargs)
		return self._View(self, **kwargs)



class AbstractCountableSource(AbstractDataSource, AbstractCountableData):
	def _validate_selection(self, naive):
		return naive



class AbstractCountableRouterView(AbstractRouterView, AbstractCountableSource):
	pass



class AbstractSelector(AbstractScope):
	def compose(self, other: 'AbstractSelector') -> 'AbstractSelector':
		raise NotImplementedError



class AbstractIndexedData(AbstractCountableSource):
	def __init__(self, *, indices=None, **kwargs):
		super().__init__(**kwargs)


	def _validate_selection(self, naive):
		if self.indices is not None:
			naive = self.indices[naive]
		return super()._validate_selection(naive)


	@property
	def size(self):
		return len(self.indices)


	@property
	def indices(self):
		raise NotImplementedError



class AbstractProgression(BatchMogul, IteratorMogul, CreativeMogul, AbstractSourcedKit, Prepared):
	def __iter__(self):
		return self


	def set_source(self, source: AbstractDataSource):
		raise NotImplementedError


	@property
	def current_batch(self) -> 'AbstractBatch':
		raise NotImplementedError


	@property
	def source(self) -> AbstractTool:
		raise NotImplementedError


	def __str__(self):
		return f'{self.__class__.__name__}({self.source})'


	def __next__(self):
		return self.create_batch()


	def create_batch(self, size=None) -> 'AbstractBatch':
		raise NotImplementedError


	def _create_context(self, source=None, **kwargs) -> AbstractContext:
		# self.prepare()
		return super()._create_context(source=self.source, **kwargs)



class AbstractBatch(AbstractCountableRouterView, AbstractSelector):
	def __init__(self, progress: AbstractProgression = None, **kwargs):
		super().__init__(**kwargs)


	def set_progress(self, progress: AbstractProgression):
		raise NotImplementedError
	@property
	def progress(self) -> AbstractProgression:
		raise NotImplementedError


	def new(self):
		return self.progress.create_batch()



class AbstractBatchable(AbstractSchema, AbstractDataSource):
	def __iter__(self):
		return self.iterate()


	def __next__(self):
		return self.batch()


	def iterate(self, batch_size: Optional[int] = unspecified_argument, **kwargs):
		raise NotImplementedError


	def batch(self, batch_size: Optional[int] = unspecified_argument, **kwargs):
		progress = self.iterate(batch_size=batch_size, **kwargs)
		return progress.current_context()






from .abstract import AbstractHyperparameter, AbstractSubmodule, AbstractBuilder, AbstractParameterized

from .building import register_builder, get_builder, BuildCreator

from .top import Builder, MultiBuilder, RegistryBuilder, RegisteredProduct, \
	Structured, HierarchyBuilder, Parameterized, Spec, Function, SimpleFunction
from .top import hparam, submodule, submachine, with_hparam, inherit_hparams, Structured, MatchingBuilder




from collections import OrderedDict#, Mapping
# import torch
from omnibelt import agnostic, unspecified_argument

from .hardware import Device
from .random import Seeded


class AbstractContainer:#(Mapping):
	pass



class Container(OrderedDict, AbstractContainer): # TODO: instead of inheriting from OrderedDict use Mapping (maybe?)
							  # http://www.kr41.net/2016/03-23-dont_inherit_python_builtin_dict_type.html
	def _find_missing(self, key):
		raise KeyError(key)


	class _missing: # flag for missing items
		pass


	def merge(self, info: 'Container'):
		self.update(info)


	def __getitem__(self, item):
		try:
			return super().__getitem__(item)
		except KeyError:
			return self._find_missing(item)
			# self[item] = val
			# return val


	def __str__(self):
		entries = ', '.join(self.keys())
		return f'{self.__class__.__name__}({entries})'


	def __repr__(self):
		return str(self)



class DevicedContainer(Device, Container):
	def _to(self, device, **kwargs):
		for key, val in self.items():
			if isinstance(val, (Device, torch.Tensor)):
				self[key] = val.to(device)



class SourceContainer(Container):
	def __init__(self, source=None, **kwargs):
		super().__init__(**kwargs)
		self.source = source


	def merge(self, info: 'Container'):
		if isinstance(info, SourceContainer) and info.source is not None:
			self.source = info.source
		return super().merge(info)


	def new_source(self, source):
		self.clear()
		self.source = source


	def _load_missing(self, key):
		return self.source[key]


	def _find_missing(self, key):
		if self.source is not None:
			self[key] = self._load_missing(key) # load and cache
			return self[key]
		return super()._find_missing(key)



class ScoreContainer(Container):
	_score_key = None
	def __init__(self, *, score_key=None, **kwargs):
		super().__init__(**kwargs)
		if score_key is not None:
			self._score_key = score_key


	def merge(self, info: 'Container'):
		if isinstance(info, ScoreContainer) and info._score_key is not None:
			self._score_key = info._score_key
		return super().merge(info)


	class NoScoreKeyError(Exception):
		pass


	def _find_missing(self, key, **kwargs):
		if key == 'score':
			if self._score_key is None:
				raise self.NoScoreKeyError
			return self[self._score_key]
		return super()._find_missing(key)


	def __contains__(self, item):
		return super().__contains__(item) or (item == 'score' and super().__contains__(self._score_key))












from typing import Optional, Any, Iterator, TypeVar, Generic, Union, Callable, Iterable, Mapping, Sequence, Self
from itertools import chain
from collections import OrderedDict
from omnibelt import filter_duplicates
from omnibelt.crafts import InheritableCrafty

from .abstract import AbstractGadget, AbstractGaggle, AbstractGig
from .errors import logger, GadgetFailure, MissingGadget, AssemblyError
from .gadgets import GadgetBase, SingleGadgetBase, FunctionGadget, AutoFunctionGadget



class GaggleBase(GadgetBase, AbstractGaggle):
	"""
	The GaggleBase class is a base class for creating custom gaggles. It uses a protected _gadgets_table dictionary to
	keep track of all the subgadgets, and consequently implements the expected API for gaggles.

	The gadgets in the table should be in O-N order (reverse order of presidence, so the last gadget in the list for
	a gizmo is tried first).

	Attributes:
		_gadgets_table (dict[str, list[AbstractGadget]]): A dictionary where keys are gadget names and values are lists
		of subgadgets.
	"""

	_gadgets_table: dict[str, list[AbstractGadget]]

	def __init__(self, *args, gadgets_table: Optional[Mapping] = None, **kwargs):
		"""
		Initializes a new instance of the GaggleBase class.

		Args:
			args: unused
			gadgets_table (Optional[Mapping]): A dictionary of gadgets. If not provided, defauls to an empty dictionary.
			kwargs: Arbitrary keyword arguments.
		"""
		if gadgets_table is None:
			gadgets_table = {}
		super().__init__(*args, **kwargs)
		self._gadgets_table = gadgets_table

	def gizmos(self) -> Iterator[str]:
		"""
		Iterates over all known gizmos that this gaggle can produce in order of oldest to newest.
		"""
		yield from self._gadgets_table.keys()

	def grabable(self, gizmo: str) -> bool:
		"""
		Checks if a gizmo is can be produced by this gaggle.

		Args:
			gizmo (str): The name of the gizmo to check.

		Returns:
			bool: True if the gizmo can be produced, False otherwise.
		"""
		return gizmo in self._gadgets_table

	def vendors(self, gizmo: Optional[str] = None) -> Iterator[AbstractGadget]:
		"""
		Returns all known subgadgets that are used by this gaggle wihtout repeats, that produces the specified gizmo,
		if provided, otherwise, iterates over all subgadgets. The iteration starts with the most recently added gadgets.

		Args:
			gizmo (Optional[str]): The name of the gizmo to check. If not provided, all gadgets are returned.

		Returns:
			Iterator[AbstractGadget]: An iterator over the gadgets that can produce the given gizmo.
		"""
		yield from self._vendors(gizmo)

	def _vendors(self, gizmo: Optional[str] = None) -> Iterator[AbstractGadget]:
		"""
		Private method that returns all known subgadgets that can produce the given gizmo.

		Args:
			gizmo (Optional[str]): The name of the gizmo to check. If not provided, all gadgets are returned.

		Returns:
			Iterator[AbstractGadget]: An iterator over the gadgets that can produce the given gizmo.
		"""
		if gizmo is None:
			for gadget in filter_duplicates(chain.from_iterable(map(reversed, self._gadgets_table.values()))):
				if isinstance(gadget, AbstractGaggle):
					yield from gadget.gadgets(gizmo)
				else:
					yield gadget
		else:
			if gizmo not in self._gadgets_table:
				raise self._MissingGadgetError(gizmo)
			for gadget in filter_duplicates(reversed(self._gadgets_table[gizmo])):
				if isinstance(gadget, AbstractGaggle):
					yield from gadget.gadgets(gizmo)
				else:
					yield gadget

	_AssemblyFailedError = AssemblyError
	def grab_from(self, ctx: AbstractGig, gizmo: str) -> Any:
		"""
		Tries to grab a gizmo using the subgadgets given the context.

		Args:
			ctx (AbstractGig): The context with which to grab the gizmo.
			gizmo (str): The name of the gizmo to grab.

		Returns:
			Any: The grabbed gizmo.

		Raises:
			AssemblyFailedError: If all subgadgets fail to produce the gizmo.
			MissingGadgetError: If no gadget can produce the gizmo.
		"""
		failures = OrderedDict()
		for gadget in self._vendors(gizmo):
			try:
				return gadget.grab_from(ctx, gizmo)
			except self._GadgetFailure as e:
				failures[e] = gadget
			except:
				logger.debug(f'{gadget!r} failed while trying to produce {gizmo!r}')
				raise
		if failures:
			raise self._AssemblyFailedError(failures)
		raise self._MissingGadgetError(gizmo)

class LoopyGaggle(GaggleBase):
	"""
	The LoopyGaggle class is mix-in for custom gaggles that allows multiple gadgets that produce the same gizmos to
	recursively call each other. Specifically, if a subgadget requires the same gizmo as input as it produces, then the
	next known subgadget of this gaggle is used after which the stack is resolved.

	Attributes:
		_grabber_stack (dict[str, Iterator[AbstractGadget]]): A dictionary keeping track of which subgadgets are still
		available to use for each gizmo.
	"""
	_grabber_stack: dict[str, Iterator[AbstractGadget]] = None

	def __init__(self, *args, **kwargs):
		"""
		Initializes a new instance of the LoopyGaggle class.

		Args:
			args: unused, passed to super.
			kwargs: unused, passed to super.
		"""
		super().__init__(*args, **kwargs)
		self._grabber_stack = {}

	def grab_from(self, ctx: 'AbstractGig', gizmo: str) -> Any:
		"""
		Tries to grab a gizmo from the context using the gadgets in the _grabber_stack dictionary.

		Args:
			ctx (AbstractGig): The context with which to grab the gizmo.
			gizmo (str): The name of the gizmo to grab.

		Returns:
			Any: The grabbed gizmo.

		Raises:
			AssemblyFailedError: If all gadgets fail to produce the gizmo.
			MissingGadgetError: If no gadget can produce the gizmo.
		"""
		failures = OrderedDict()
		itr = self._grabber_stack.setdefault(gizmo, self._vendors(gizmo))
		for gadget in itr:
			try:
				out = gadget.grab_from(ctx, gizmo)
			except self._GadgetFailure as e:
				failures[e] = gadget
			except:
				logger.debug(f'{gadget!r} failed while trying to produce {gizmo!r}')
				raise
			else:
				if gizmo in self._grabber_stack:
					self._grabber_stack.pop(gizmo)
				return out
		if gizmo in self._grabber_stack:
			self._grabber_stack.pop(gizmo)
		if failures:
			raise self._AssemblyFailedError(failures)
		raise self._MissingGadgetError(gizmo)

class MutableGaggle(GaggleBase):
	"""
	The MutableGaggle class is a mix-in for custom gaggles to dynamically add and remove subgadgets.
	"""

	def include(self, *gadgets: AbstractGadget) -> Self:
		"""
		Adds given gadgets in the order that is given, which means subsequent `grab` would use the first provided
		gadget before trying the next.

		Args:
			gadgets (AbstractGadget): The gadgets to be added.

		Returns:
			Self: this gaggle.
		"""
		return self.extend(gadgets)

	def extend(self, gadgets: Iterable[AbstractGadget]) -> Self:
		"""
		Adds given gadgets in the iterator in the order that is given, which means subsequent `grab` would use the
		first provided gadget before trying the next.

		Args:
			gadgets (Iterable[AbstractGadget]): The gadgets to be added.

		Returns:
			Self: this gaggle.
		"""
		new = {}
		for gadget in gadgets:
			for gizmo in gadget.gizmos():
				new.setdefault(gizmo, []).append(gadget)
		for gizmo, gadgets in new.items():
			if gizmo in self._gadgets_table:
				for gadget in gadgets:
					if gadget in self._gadgets_table[gizmo]:
						self._gadgets_table[gizmo].remove(gadget)
			self._gadgets_table.setdefault(gizmo, []).extend(reversed(gadgets))
		return self

	def exclude(self, *gadgets: AbstractGadget) -> Self:
		"""
		Removes the given gadgets, if they are found.

		Args:
			gadgets (AbstractGadget): The gadgets to be removed.

		Returns:
			Self: this gaggle.
		"""
		for gadget in gadgets:
			for gizmo in gadget.gizmos():
				if gizmo in self._gadgets_table and gadget in self._gadgets_table[gizmo]:
					self._gadgets_table[gizmo].remove(gadget)
		return self

class CraftyGaggle(GaggleBase, InheritableCrafty):
	"""
	The CraftyGaggle class is a mix-in for custom gaggles to handle crafts such as `tool`.

	Note that in order to procuess and add all found crafts, a subclass must call `_process_crafts()`, which is not
	done here.
	"""

	def _process_crafts(self):
		"""
		Identifies all crafts contained in the class (including super classes) that are gadgets and adds those as
		subgadgets, in the order in which they are defined.

		Note that, in order to be processed correctly, the crafts should produce skills which are instances of
		`AbstractGadget`.
		"""
		# group by where the craft is defined
		history = OrderedDict() # src<N-O> : craft<N-O>
		for src, key, craft in self._emit_all_craft_items(): # craft<N-O>
			history.setdefault(src, []).append(craft)

		# convert crafts to skills and add in O-N (N-O) order to table
		for crafts in reversed(history.values()): # O-N
			gizmos = {}
			for craft in reversed(crafts): # N-O (in order of presidence)
				skill = craft.as_skill(self)
				if isinstance(skill, AbstractGadget):
					for gizmo in skill.gizmos():
						gizmos.setdefault(gizmo, []).append(skill)
			for gizmo, skills in gizmos.items():
				self._gadgets_table.setdefault(gizmo, []).extend(reversed(skills)) # O-N (in order of appearance)





from .operations import Extractor, Encoder, Decoder, Generator, Discriminator, Augmentation, Criterion, \
	Metric, PathCriterion, Interpolator, Estimator, Invertible, Compressor, Quantizer, Sampler
from .metrics import Measure, Norm, Lp, L0, L1, L2, Linf
from .functions import Function
from . import spaces
from .distributions import Distribution, DistributionTensor, NormalDistribution


from typing import Union, Any, Callable, Type, Iterable, Iterator, Optional, List, Dict, Tuple, Sequence, Hashable
from pathlib import Path
import numpy as np
# import torch
from omnibelt import unspecified_argument, agnosticmethod, md5, primitive, JSONABLE
from omnibelt import Exportable, ExportManager

from .errors import ExtractionError, UnknownObjectError, NoObjectError


# TODO: make exportable
# .fp -> hash
# .fpd -> data


class AbstractFingerprinted:
	@property
	def fingerprint(self):
		raise NotImplementedError



class AbstractFingerprint(AbstractFingerprinted, Exportable):
	_ExtractionError = ExtractionError
	_UnknownObjectError = UnknownObjectError
	_NoObjectError = NoObjectError


	def __init__(self, src=None, **kwargs):
		super().__init__(**kwargs)


	@property
	def fingerprint(self):
		return self


	@classmethod
	def extract(cls, obj, force_str=False) -> JSONABLE:
		if force_str:
			if not isinstance(obj, str):
				raise cls._ExtractionError(obj)
			return str(obj)
		if isinstance(obj, primitive):
			return obj
		# if isinstance(obj, (np.ndarray, torch.Tensor)):
		# 	numels = np.product(obj.shape).item()
		# 	sel = torch.randint(numels, size=(min(5, numels),),
		# 	        generator=torch.Generator().manual_seed(16283393149723337453)).tolist()
		# 	return [str(obj.shape), str(obj.dtype), obj.sum().item(), obj.reshape(-1)[sel].tolist()]
		if isinstance(obj, (list, tuple)):
			return [cls.extract(o) for o in obj]
		if isinstance(obj, dict):
			return {cls.extract(k, force_str=True): cls.extract(v) for k, v in obj.items()}
		raise cls._ExtractionError(obj)


	def __eq__(self, other):
		return isinstance(other, AbstractFingerprint) and self.check_fingerprint(other)


	def __hash__(self):
		return hash(self.code())


	def code(self):
		return md5(self.data())


	def data(self):
		raise NotImplementedError


	def check_fingerprint(self, obj: Union['AbstractFingerprint', AbstractFingerprinted]):
		if isinstance(obj, AbstractFingerprinted):
			return self == obj.fingerprint
		if not isinstance(obj, AbstractFingerprint):
			raise self._UnknownObjectError(obj)
		return self.code() == obj.code()


	def __str__(self):
		return f'{self.__class__.__name__}({self.code()})'



class Fingerprint(AbstractFingerprint, extensions=['fp', 'fpd']):
	def __init__(self, src=None, *, data=None, code=None, **kwargs):
		super().__init__(src=src, **kwargs)
		self.src = src
		self._data = data
		self._code = code


	@classmethod
	def _load_export(cls, path: Path, src: Type['ExportManager']) -> Any:
		if path.suffix == '.fpd':
			return cls(data=src.load_export(fmt='json', path=path))
		return cls(code=path.read_text())


	@staticmethod
	def _export_payload(payload: 'AbstractFingerprint', path: Path, src: Type['ExportManager']) -> Optional[Path]:
		if path.suffix == '.fpd':
			return src.export(payload.data(), fmt='json', path=path, sort_keys=True)
		path.write_text(payload.code())
		return path


	@classmethod
	def extract(cls, obj, **kwargs) -> JSONABLE:
		if isinstance(obj, Fingerprinted):
			return cls.extract(obj._fingerprint_data(), **kwargs)
		return super().extract(obj, **kwargs)


	@property
	def src(self):
		src = getattr(self, '_src', None)
		if src is None:
			raise self._NoObjectError(src)
		return src
	@src.setter
	def src(self, src):
		self._src = src


	def data(self):
		if self._data is None:
			self._data = self.extract(self.src)
		return self._data


	def code(self):
		if self._code is None:
			self._code = super().code()
		return self._code



class Fingerprinted(AbstractFingerprinted):
	_Fingerprint = Fingerprint


	@property
	def fingerprint(self):
		return self._Fingerprint(self)


	def _fingerprint_data(self):
		return {'cls': self.__class__.__name__, 'module': self.__module__}









from typing import Iterator, Optional, Any, Iterable, Callable
from omnibelt.crafts import AbstractSkill, AbstractCraft, AbstractCrafty, NestableCraft

from .abstract import AbstractGadget, AbstractGaggle, AbstractGig
from .gadgets import GadgetBase, SingleGadgetBase, FunctionGadget, AutoFunctionGadget


class ToolSkill(AbstractSkill):
	"""
	The ToolSkill class is a subclass of AbstractSkill. It provides methods to handle unbound functions and their bases.

	Attributes:
		_unbound_fn (Callable): The unbound function to be handled.
		_base (Optional[AbstractCraft]): The base of the unbound function. Defaults to None.
	"""

	def __init__(self, *, unbound_fn: Callable, base: Optional[AbstractCraft] = None, **kwargs):
		"""
		Initializes a new instance of the ToolSkill class.

		Args:
			unbound_fn (Callable): The unbound function to be handled.
			base (Optional[AbstractCraft]): The base of the unbound function. If not provided, base will be None.
			kwargs: Arbitrary keyword arguments.
		"""
		super().__init__(**kwargs)
		self._unbound_fn = unbound_fn
		self._base = base

	def __get__(self, instance, owner):
		"""
		Returns the unbound function if the instance is None, otherwise it returns the bound method.

		Args:
			instance: The instance that the method is being accessed through, or None when the method is accessed through the owner.
			owner: The owner class.

		Returns:
			Callable: The unbound function if the instance is None, otherwise the bound method.
		"""
		if instance is None:
			return self
		return self._unbound_fn.__get__(instance, owner)

class ToolCraft(FunctionGadget, NestableCraft):
	"""
	The ToolCraft class is a subclass of FunctionGadget and NestableCraft. It provides methods to handle gizmos and their associated functions.

	Attributes:
		_ToolSkill (FunctionGadget, ToolSkill): A nested class that inherits from FunctionGadget and ToolSkill.
	"""

	@property
	def __call__(self):
		"""
		Calling a ToolCraft instance directly results in the wrapped function being called.

		Returns:
			Callable: The wrapped function.
		"""
		return self._wrapped_content_leaf()

	def __get__(self, instance, owner):
		"""
		When accessing ToolCraft instances directly, they behave as regular methods, applying __get__ to the wrapped function.

		Args:
			instance: The instance that the method is being accessed through, or None when the method is accessed through the owner.
			owner: The owner class.

		Returns:
			Callable: The wrapped function.
		"""
		return self._wrapped_content_leaf().__get__(instance, owner)

	def _wrapped_content(self):
		"""
		Returns the wrapped function. This may be a nested craft or other decorator.

		Note: If you want the actual function, use _wrapped_content_leaf.

		Returns:
			Callable: The wrapped function.
		"""
		return self._fn

	class _ToolSkill(FunctionGadget, ToolSkill):
		"""
		The _ToolSkill class is a nested class that inherits from FunctionGadget and ToolSkill.
		"""
		pass

	def as_skill(self, owner: AbstractCrafty) -> ToolSkill:
		"""
		When an AbstractCrafty is instantiated (i.e., `owner`), any crafts accessible by the class (including inherited ones) can be converted to skills.

		Args:
			owner (AbstractCrafty): The owner of the craft.

		Returns:
			ToolSkill: The converted skill.
		"""
		unbound_fn = self._wrapped_content_leaf()
		fn = unbound_fn.__get__(owner, type(owner))
		return self._ToolSkill(self._gizmo, fn=fn, unbound_fn=unbound_fn, base=self)


class AutoToolCraft(AutoFunctionGadget, ToolCraft):
	"""
	The AutoToolCraft class is a subclass of AutoFunctionGadget and ToolCraft. It provides methods to handle automatic
	function gadgets and tool crafts.

	Attributes:
		_ToolSkill (AutoFunctionGadget, ToolSkill): A nested class that inherits from AutoFunctionGadget and ToolSkill.
	"""

	class _ToolSkill(AutoFunctionGadget, ToolSkill):
		"""
		The _ToolSkill class is a nested class that inherits from AutoFunctionGadget and ToolSkill. It provides methods
		to handle automatic function gadgets and tool skills.
		"""
		pass

class ToolDecorator(GadgetBase):
	"""
	The ToolDecorator class is a subclass of GadgetBase. It provides methods to handle gizmo decoration.

	Attributes:
		_gizmo_type (None): The type of the gizmo. Defaults to None.
		_gizmo (str): The gizmo to be handled.
		_ToolCraft (ToolCraft): A nested class that inherits from ToolCraft.
	"""

	_gizmo_type = None

	def __init__(self, gizmo: str, **kwargs):
		"""
		Initializes a new instance of the ToolDecorator class.

		Args:
			gizmo (str): The gizmo to be handled.
			kwargs: Arbitrary keyword arguments.
		"""
		if isinstance(gizmo, str) and self._gizmo_type is not None:
			gizmo = self._gizmo_type(gizmo)
		super().__init__(**kwargs)
		self._gizmo = gizmo

	def gizmos(self) -> Iterator[str]:
		"""
		Lists gizmos produced by self.

		Returns:
			Iterator[str]: An iterator over the gizmos.
		"""
		yield self._gizmo

	def grabable(self, gizmo: str) -> bool:
		"""
		Checks if a gizmo is grabable.

		Args:
			gizmo (str): The name of the gizmo to check.

		Returns:
			bool: True if the gizmo is grabable, False otherwise.
		"""
		return gizmo == self._gizmo

	def grab_from(self, ctx: Optional[AbstractGig], gizmo: str) -> Any:
		"""
		Tries to grab a gizmo from the context.

		Args:
			ctx (Optional[AbstractGig]): The context from which to grab the gizmo.
			gizmo (str): The name of the gizmo to grab.

		Returns:
			Any: The grabbed gizmo.

		Raises:
			_GadgetFailed: If the gizmo cannot be grabbed.
		"""
		raise self._GadgetFailed(gizmo)

	_ToolCraft = ToolCraft

	def _actualize_tool(self, fn: Callable, **kwargs):
		"""
		Actualizes a tool by creating a ToolCraft instance with the gizmo and the function.

		Args:
			fn (Callable): The function to be actualized.
			kwargs: Arbitrary keyword arguments.

		Returns:
			ToolCraft: The actualized tool.
		"""
		return self._ToolCraft(self._gizmo, fn=fn, **kwargs)

	def __call__(self, fn):
		"""
		Calling a ToolDecorator instance directly results in the actualization of the function.

		Args:
			fn (Callable): The function to be actualized.

		Returns:
			ToolCraft: The actualized tool.
		"""
		return self._actualize_tool(fn)

class AutoToolDecorator(ToolDecorator):
	"""
	The AutoToolDecorator class is a subclass of ToolDecorator. It overrides the _ToolCraft attribute of the parent class
	with AutoToolCraft. This means that when a tool is actualized, an instance of AutoToolCraft will be created instead
	of ToolCraft.

	Attributes:
		_ToolCraft (AutoToolCraft): A nested class that inherits from AutoToolCraft.
	"""
	_ToolCraft = AutoToolCraft










from collections import OrderedDict
# import torch
# from torch import nn
from omnibelt import agnostic, unspecified_argument#, mix_into
# from .. import util

from ..parameters import submodule, Structured
from ..features import Prepared

from .abstract import AbstractModel, AbstractTrainableModel, AbstractTrainer
from .simple import Evaluatable



class Trainer(Structured, AbstractTrainer):
	model = submodule(builder='model')

	def __init__(self, model=None, **kwargs):
		super().__init__(model=model, **kwargs)
		self.model = model
		self._num_iter = 0


	def _prepare(self, *args, **kwargs):
		super()._prepare(*args, **kwargs)
		self.model.prepare(*args, **kwargs)


	def loop(self, source, **kwargs):
		self.loader = source.get_iterator(**kwargs)
		for batch in self.loader:
			info = self.create_step_container(source=batch, **kwargs)
			yield info


	def create_step_container(self, *args, **kwargs):
		return self.model.create_step_container(*args, **kwargs)


	def fit(self, source, **kwargs):
		self.prepare(source=source)
		info = None
		for batch in self.loop(source, **kwargs):
			info = self.step(batch)
		return self.finish_fit(info)


	def evaluate(self, source, **kwargs):
		info = self.model.evaluate(source, **kwargs)
		return self.finish_evaluate(info)


	def step(self, info):
		out = self.model.step(info)
		self._num_iter += 1
		return out



class TrainableModel(Evaluatable, Structured, AbstractTrainableModel):
	_Trainer = Trainer



# class Loggable(Model): # TODO: this should be in the trainer! the Model just has a function
# # 	def __init__(self, stats=None, **kwargs):
# # 		if stats is None:
# # 			stats = self.Statistics()
# # 		super().__init__(**kwargs)
# # 		self._stats = stats
# #
# #
# # 	class Statistics:
# # 		def mete(self, info, **kwargs):
# # 			raise NotImplementedError
# #
# #
# # 	# def register_stats(self, *stats):
# # 	# 	for stat in stats:
# # 	# 		self._stats.append(stat)
# # 	# 	# for name in names:
# # 	# 	# 	self._stats[name] = self.Statistic(name)
# # 	# 	# for name, stat in stats.items():
# # 	# 	# 	if not isinstance(stat, self.Statistic):
# # 	# 	# 		print(f'WARNING: stat {name} should subclass {self.Statistic.__name__}')
# # 	# 	# 	self._stats[name] = stat
# #
# #
# 	def log(self, info, **kwargs):
# 		pass
# 		# self._stats.mete(info, **kwargs)




# Types of Models








import torch as _torch
from .features import Device as _Device

# TODO: separate with defaults
from omniplex.framework.common import \
	Extractor as _Extractor, \
	Encoder as _Encoder, \
	Decoder as _Decoder, \
	Generator as _Generator, \
	Discriminator as _Discriminator, \
	Augmentation as _Augmentation, \
	Criterion as _Criterion, \
	Metric as _Metric, \
	PathCriterion as _PathCriterion, \
	Interpolator as _Interpolator, \
	Estimator as _Estimator, \
	Invertible as _Invertible, \
	Compressor as _Compressor, \
	Quantizer as _Quantizer, \
	Function as _Function



class _ImplicitModel(_Function, _Device):
	def __init__(self, fn, din_device=None, dout_device=None, **kwargs):
		super().__init__(**kwargs)
		self.fn = fn
		self.din_device = din_device
		self.dout_device = dout_device


	def _wrapped_call(self, *args):
		return self._process_output(self.fn(*map(self._process_input, args)))


	def _process_input(self, inp):
		device = self.din_device
		if device is None:
			device = self.device
		if device is not None and isinstance(inp, (_torch.Tensor, _Device)):
			inp = inp.to(device)
		return inp


	def _process_output(self, out):
		device = self.dout_device
		if device is None:
			device = self.device
		if device is not None and isinstance(out, (_torch.Tensor, _Device)):
			out = out.to(device)
		return out



class _ImplicitInvertibleModel(_ImplicitModel):
	def __init__(self, fn, ifn=None, **kwargs):
		super().__init__(fn, **kwargs)
		self.ifn = ifn


	def _wrapped_inv_call(self, *args):
		return self._process_output(self.fn(*map(self._process_input, args)))



class Extractor(_Extractor, _ImplicitModel):
	def extract(self, observation):
		return self._wrapped_call(observation)



class Encoder(_Encoder, Extractor):
	def encode(self, observation):
		return self._wrapped_call(observation)



class Decoder(_Decoder, _ImplicitModel):
	def decode(self, latent):
		return self._wrapped_call(latent)



class Generator(_Generator, _ImplicitModel):
	def sample(self, *shape, gen=None):
		return self._process_output(self.fn(*shape, gen=gen))



class Discriminator(_Discriminator, _ImplicitModel):
	def judge(self, observation):
		return self._wrapped_call(observation)



class Criterion(_Criterion, _ImplicitModel):
	def compare(self, observation1, observation2):
		return self._wrapped_call(observation1, observation2)



class Metric(_Metric, Criterion): # obeys triangle inequality
	def distance(self, observation1, observation2):
		return self._wrapped_call(observation1, observation2)



class Interpolator(_Interpolator, _ImplicitModel):
	def interpolate(self, start, end, N):
		return self._process_output(self.fn(self._process_input(start), self._process_input(end), N))



class Estimator(_Estimator, _ImplicitModel):
	def predict(self, observation):
		return self._wrapped_call(observation)



class Augmentation(_Augmentation, _ImplicitModel):
	def augment(self, observation):
		return self._wrapped_call(observation)



class PathCriterion(_PathCriterion, Criterion):
	def compare_path(self, path1, path2):
		return self._wrapped_call(path1, path2)



class Invertible(_Invertible, _ImplicitInvertibleModel):
	def forward(self, observation):
		return self._wrapped_call(observation)


	def inverse(self, observation):
		return self._wrapped_inv_call(observation)



class Compressor(_Compressor, _ImplicitInvertibleModel):
	def compress(self, observation):
		return self.fn(self._process_input(observation))


	def decompress(self, data):
		return self._process_output(self.ifn(data))



class Quantizer(_Quantizer, _ImplicitInvertibleModel):
	def quantize(self, observation):
		return self._wrapped_call(observation)


	def dequantize(self, observation):
		return self._wrapped_inv_call(observation)












from typing import Tuple, List, Dict, Optional, Union, Any, Callable, Sequence, Iterator, Iterable, Type, Set
import math
# from ..persistent import Fingerprinted
from ..tools import Industrial
from ..parameters import Parameterized, Structured, hparam

from .abstract import AbstractDataSource
from .routers import DataCollection, AutoCollection, AliasedCollection, CountableDataRouter
from .views import IndexSelector
from .sources import Splitable, TensorSource, SpacedSource#, BuildableData
from .progression import SetProgression, StreamProgression, AbstractProgression
from .budgeting import BudgetLoader
from .batches import Batchable, BatchBase, IndexBatch



class Bunch(BatchBase):
	pass
StreamProgression._Context = Bunch



class Batch(IndexBatch):
	_Progression = SetProgression
	pass
SetProgression._Context = Batch



class Buffer(Parameterized, Batchable, TensorSource, SpacedSource):#, BuildableData): # TODO: should be epochable
	_Progression = SetProgression
	pass



# print()
# print('\n'.join(map(str, Structured.mro())))
# print()
# print('\n'.join(map(str, Batchable.mro())))
# print()
# print('\n'.join(map(str, AutoCollection.mro())))
# print()
# print('\n'.join(map(str, AliasedCollection.mro())))
# print()
# print('\n'.join(map(str, DataCollection.mro())))
class _FeaturedDataRouter(AutoCollection, AliasedCollection, DataCollection, Structured, Batchable):
	# #, BuildableData):
	# def register_buffer(self, gizmo: str, buffer=None, **kwargs):
	# 	buffer = super().register_buffer(gizmo, buffer, **kwargs)
	# 	if gizmo in self._spaces:
	# 		del self._spaces[gizmo]
	# 	return buffer
	pass



class Datastream(_FeaturedDataRouter): # not countable (but batchable)
	_Progression = StreamProgression
	pass



class Subset(Splitable, Batchable, IndexSelector):
	_Progression = SetProgression
	pass



class Dataset(Splitable, CountableDataRouter, _FeaturedDataRouter):
	_Progression = SetProgression
	_Buffer = Buffer
	_Subset = Subset



class SimpleDataset(Dataset):
	_is_ready = True

	def __init__(self, *unnamed_data, **named_data):
		super().__init__()
		self._register_init_data(unnamed_data, named_data)


	def _register_init_data(self, unnamed_data, named_data):
		for i, x in enumerate(unnamed_data):
			self.register_buffer(i, x)
		for k, v in named_data.items():
			self.register_buffer(k, v)



######










from typing import Tuple, List, Dict, Optional, Union, Any, Callable, Sequence, Iterator, Iterable

from collections import OrderedDict
from omnibelt import unspecified_argument, get_printer

import math
# import torch

from ..structure import Generator, Sampler
from .. import util
from ..features import Seedable
from ..structure import spaces
# from ..parameters import Buildable
from .abstract import AbstractDataRouter, AbstractDataSource, \
	AbstractSelector, AbstractBatchable, AbstractCountableData, AbstractCountableSource
from .views import SizeSelector, IndexSelector

prt = get_printer(__file__)


# class BuildableData(Buildable, AbstractDataSource):
# 	pass



class Shufflable(Seedable):
	@staticmethod
	def _is_big_number(N):
		return N > 20000000


	def _shuffle_indices(self, N):
		# TODO: include a warning if cls._is_big_number(N)
		return torch.randint(N, size=(N,), generator=self.rng) \
			if self._is_big_number(N) else torch.randperm(N, generator=self.rng)



# class BatchableSource(AbstractBatchable):
# 	_Batch = None



class SpacedSource(AbstractDataSource):
	def __init__(self, *, space=None, **kwargs):
		super().__init__(**kwargs)
		self._space = space


	@property
	def space(self):
		return self._space
	@space.setter
	def space(self, space):
		self._space = space



# class SingleSource(AbstractDataSource):
# 	def _get_from(self, source, key=None):
# 		return self._get(source)
#
# 	@staticmethod
# 	def _get(source):
# 		raise NotImplementedError



# class SampleSource(AbstractDataSource, Sampler):
# 	_sample_key = None
# 	def _sample(self, shape, *, sample_key=None):
# 		if sample_key is unspecified_argument:
# 			sample_key = self._sample_key
# 		N = shape.numel()
# 		samples = self.sample_buffer(sample_key, N)
# 		return util.split_dim(samples, *shape)
#
#
# 	def sample_buffer(self, N, key):
# 		return self.get_from(N, key)



# class SelectorSource(AbstractDataSource, AbstractSelector):
# 	def get(self, key):
# 		return self.get_from(self, key)



# class GenerativeSource(AbstractDataSource, Generator):
# 	def _get_from(self, source, key):
# 		return self.generate(source.size)



class Subsetable(AbstractCountableSource, Shufflable):
	@staticmethod
	def _split_indices(indices, cut):
		assert cut != 0
		last = cut < 0
		cut = abs(cut)
		total = len(indices)
		if isinstance(cut, float):
			assert 0 < cut < 1
			cut = int(cut * total)
		part1, part2 = indices[:cut], indices[cut:]
		if last:
			part1, part2 = part2, part1
		return part1, part2
	
	_Subset = None # indexed view
	def subset(self, cut=None, *, indices=None, shuffle=False):
		# if not hard_copy:
		# 	raise NotImplementedError # TODO: hard copy
		if indices is None:
			assert cut is not None, 'Either cut or indices must be specified'
			indices, _ = self._split_indices(indices=self._shuffle_indices(self.size) \
				if shuffle else torch.arange(self.size), cut=cut)
		indices = self._validate_selection(indices)
		return self._Subset(source=self, indices=indices)
	


class Splitable(Subsetable):
	def split(self, splits, shuffle=False):
		auto_name = isinstance(splits, (list, tuple, set))
		if auto_name:
			named_cuts = [(f'part{i}', r) for i, r in enumerate(splits)]
		else:
			assert isinstance(splits, dict), f'unknown splits: {splits}'
			assert not any(x for x in splits if x is None), 'names of splits cannot be None'
			named_cuts = list(splits.items())
		names, cuts = zip(*sorted(named_cuts, key=lambda nr: (isinstance(nr[1], int), isinstance(nr[1], float),
		                                                      nr[1] is None, nr[0]), reverse=True))

		remaining = self.size
		nums = []
		itr = iter(cuts)
		for cut in itr:
			if isinstance(cut, int):
				nums.append(cut)
				remaining -= cut
			else:
				if isinstance(cut, float):
					ratios = []
					while isinstance(cut, float):
						ratios.append(cut)
						cut = next(itr, 'done')
					if len(cuts):
						rationums = [int(remaining * abs(ratio)) for ratio in ratios]
						nums.extend([int(math.copysign(1, r) * n) for r, n in zip(ratios, rationums)])
						remaining -= sum(rationums)
				if cut is None:
					pieces = len([cut, *itr])
					assert remaining > pieces, f'cant evenly distribute {remaining} samples into {pieces} cuts'
					evennums = [int(remaining // pieces) for _ in range(pieces)]
					nums.extend(evennums)
					remaining -= sum(evennums)

		if remaining > 0:
			nums[-1] += remaining

		indices = self._shuffle_indices(self.size) if shuffle else torch.arange(self.size)

		plan = dict(zip(names, nums))
		parts = {}
		for name in sorted(names):
			num = plan[name]
			part, indices = self._split_indices(indices, num)
			parts[name] = self.subset(indices=part)
		if auto_name:
			return [parts[name] for name, _ in named_cuts]
		return parts



class TensorSource(AbstractCountableData, AbstractDataSource): # TODO: make batchable
	def __init__(self, data=None, space=None, **kwargs):
		super().__init__(**kwargs)
		self._data = data
		self._space = space


	@property
	def is_ready(self):
		return self._data is not None


	@property
	def space(self):
		return self._space
	@space.setter
	def space(self, space):
		self._space = space


	def space_of(self, gizmo: str) -> spaces.Dim:
		return self.space


	@property
	def size(self):
		return len(self._data)


	@property
	def data(self):
		return self._data
	@data.setter
	def data(self, data):
		self._data = data


	def gizmos(self) -> Iterator[str]:
		yield from ()


	def get_from(self, source, gizmo=None):
		if source is None or not isinstance(source, IndexSelector):
			return self._data[:source.size] if isinstance(source, SizeSelector) else self._data
		return self.data[source.indices]



class MultiModed(Splitable): # TODO: splitting datasets into modes
	pass








# class ProcessDataset(Dataset):
#
# 	process = None # TODO: maybe machine?
#
# 	def _prepare(self, *args, **kwargs):
# 		raise NotImplementedError
#
#
#
# class RootedDataset(DataCollection, Rooted):
# 	@classmethod
# 	def _infer_root(cls, root=None):
# 		return super()._infer_root(root=root) / 'datasets'
#
#
# 	@agnostic
# 	def get_root(self, dataset_dir=None):
# 		if dataset_dir is None:
# 			dataset_dir = self.name
# 		root = super().get_root() / dataset_dir
# 		os.makedirs(str(root), exist_ok=True)
# 		return root
#
#
# 	def get_aux_root(self, dataset_dir=None):
# 		root = self.get_root(dataset_dir=dataset_dir) / 'aux'
# 		os.makedirs(str(root), exist_ok=True)
# 		return root
#
#
# 	def _find_path(self, dataset_name='', file_name=None, root=None):
# 		if root is None:
# 			root = self.root
# 		*other, dataset_name = dataset_name.split('.')
# 		if file_name is None:
# 			file_name = '.'.join(other) if len(other) else self.name
# 		path = root / f'{file_name}.h5'
# 		return path, dataset_name
#
#
# 	_default_hdf_buffer_type = HDFBuffer
# 	def register_hdf_buffer(self, name, dataset_name, file_name=None, root=None,
# 	                        buffer_type=None, path=None, **kwargs):
# 		if buffer_type is None:
# 			buffer_type = self._default_hdf_buffer_type
# 		if path is None:
# 			path, dataset_name = self._find_path(dataset_name, file_name=file_name, root=root)
# 		return self.register_buffer(name, buffer_type=buffer_type, dataset_name=dataset_name, path=path, **kwargs)
#
#
# 	@staticmethod
# 	def create_hdf_dataset(path, dataset_name, data=None, meta=None, dtype=None, shape=None):
# 		# if file_name is unspecified_argument:
# 		# 	file_name = 'aux'
# 		# if path is None:
# 		# 	path, dataset_name = self._find_path(dataset_name, file_name=file_name, root=root)
#
# 		if isinstance(data, torch.Tensor):
# 			data = data.detach().cpu().numpy()
# 		with hf.File(path, 'a') as f:
# 			if data is not None or (dtype is not None and shape is not None):
# 				f.create_dataset(dataset_name, data=data, dtype=dtype, shape=shape)
# 			if meta is not None:
# 				f.attrs[dataset_name] = json.dumps(meta, sort_keys=True)
# 		return path, dataset_name
#
#
#
# class DownloadableDataset(RootedDataset):
# 	def __init__(self, download=False, **kwargs):
# 		super().__init__(**kwargs)
# 		self._auto_download = download
#
#
# 	@classmethod
# 	def download(cls, **kwargs):
# 		raise NotImplementedError
#
#
# 	class DatasetNotDownloaded(FileNotFoundError):
# 		def __init__(self):
# 			super().__init__('use download=True to enable automatic download.')
#
#
#
# class EncodableDataset(ObservationDataset, RootedDataset):
# 	def __init__(self, encoder=None, replace_observation_key=None, encoded_key='encoded',
# 	             encoded_file_name='aux', encode_on_load=False, save_encoded=False, encode_pbar=None, **kwargs):
# 		super().__init__(**kwargs)
# 		self._replace_observation_key = replace_observation_key
# 		self._encoded_observation_key = encoded_key
# 		self._encoded_file_name = encoded_file_name
# 		self._encode_on_load = encode_on_load
# 		self._save_encoded = save_encoded
# 		self._encode_pbar = encode_pbar
# 		self.encoder = encoder
#
# 	@property
# 	def encoder(self): # TODO: make this a machine
# 		return self._encoder
# 	@encoder.setter
# 	def encoder(self, encoder):
# 		buffer = self.get_buffer(self._encoded_observation_key)
# 		if buffer is not None:
# 			buffer.encoder = encoder
# 		self._encoder = encoder
#
#
# 	def _get_code_path(self, file_name='aux', root=None):
# 		return None if file_name is None else self._find_path(file_name=file_name, root=root)[0]
#
#
# 	@staticmethod
# 	def _encoder_save_key(encoder):
# 		info = encoder.get_encoder_fingerprint()
# 		ident = md5(json.dumps(info, sort_keys=True))
# 		return ident, info
#
#
# 	def load_encoded_data(self, encoder=None, source_key='observation',
# 	                      batch_size=None, save_encoded=None,
# 	                      file_name=unspecified_argument, root=None):
# 		if encoder is None:
# 			encoder = self.encoder
# 		if file_name is unspecified_argument:
# 			file_name = self._encoded_file_name
# 		if save_encoded is None:
# 			save_encoded = self._save_encoded
# 		data = None
#
# 		path = self._get_code_path(file_name=file_name, root=root)
# 		if path is not None and path.exists() and encoder is not None:
# 			ident, _ = self._encoder_save_key(encoder)
# 			with hf.File(str(path), 'r') as f:
# 				if ident in f:
# 					data = f[ident][()]
# 					data = torch.from_numpy(data)
#
# 		if data is None and self._encode_on_load:
# 			batches = []
# 			for batch in self.get_iterator(batch_size=batch_size, shuffle=False, force_batch_size=False,
# 			                               pbar=self._encode_pbar):
# 				with torch.no_grad():
# 					batches.append(encoder(batch[source_key]))
#
# 			data = torch.cat(batches)
# 			if save_encoded:
# 				self.save_encoded_data(encoder, data, path)
#
# 		return data
#
#
# 	@classmethod
# 	def save_encoded_data(cls, encoder, data, path):
# 		ident, info = cls._encoder_save_key(encoder)
# 		cls.create_hdf_dataset(path, ident, data=data, meta=info)
#
#
# 	def _prepare(self, *args, **kwargs):
# 		super()._prepare(*args, **kwargs)
#
# 		if self._replace_observation_key is not None:
# 			self._encoded_observation_key = 'observation'
# 			self.register_buffer(self._replace_observation_key, self.get_buffer('observation'))
#
# 		self.register_buffer(self._encoded_observation_key,
# 		                     buffer=self.EncodedBuffer(encoder=self.encoder, source=self.get_buffer('observation'),
# 		                                               data=self.load_encoded_data()))
#
#
# 	class EncodedBuffer(BufferView):
# 		def __init__(self, encoder=None, max_batch_size=64, pbar=None, pbar_desc='encoding', **kwargs):
# 			super().__init__(**kwargs)
# 			# if encoder is not None and encoder_device is None:
# 			# 	encoder_device = getattr(encoder, 'device', None)
# 			# self._encoder_device = encoder_device
# 			self.encoder = encoder
# 			self.pbar = pbar
# 			self.pbar_desc = pbar_desc
# 			self.max_batch_size = max_batch_size
#
#
# 		@property
# 		def encoder(self):
# 			return self._encoder
# 		@encoder.setter
# 		def encoder(self, encoder):
# 			self._encoder = encoder
# 			if encoder is not None and hasattr(encoder, 'dout'):
# 				self.space = getattr(encoder, 'dout', None)
# 		# self._encoder_device = getattr(encoder, 'device', self._encoder_device)
#
#
# 		def _encode_raw_observations(self, observations):
# 			# device = observations.device
# 			if len(observations) > self.max_batch_size:
# 				samples = []
# 				batches = observations.split(self.max_batch_size)
# 				if self.pbar is not None:
# 					batches = self.pbar(batches, desc=self.pbar_desc)
# 				for batch in batches:
# 					# with torch.no_grad():
# 					# if self._encoder_device is not None:
# 					# 	batch = batch.to(self._encoder_device)
# 					samples.append(self.encoder.encode(batch)  )  # .to(device))
# 				return torch.cat(samples)
# 			# with torch.no_grad():
# 			# if self._encoder_device is not None:
# 			# 	observations = observations.to(self._encoder_device)
# 			return self.encoder.encode(observations  )  # .to(device)
#
#
# 		def _get(self, sel=None, **kwargs):
# 			sample = super()._get(sel=sel, **kwargs)
# 			if self.data is None and self.encoder is not None:
# 				sample = self._encode_raw_observations(sample)
# 			if sel is None:
# 				self.data = sample
# 			return sample
#
#
#
# class ImageDataset(ObservationDataset):
# 	class ImageBuffer(Buffer):
# 		def process_image(self, image):
# 			if not self.space.as_bytes:
# 				return image.float().div(255)
# 			return image
#
#
# 		def _get(self, *args, **kwargs):
# 			return self.process_image(super()._get(*args, **kwargs))



















from ..features.containers import SourceContainer, ScoreContainer
from ..parameters import Structured

from .abstract import AbstractFitable, AbstractEvaluatable, AbstractModel



# class Resultable(AbstractResultable):
# 	class DataContainer(SourceContainer, ScoreContainer):
# 		pass



class Evaluatable(AbstractEvaluatable):
	def create_eval_container(self, *args, **kwargs):
		return self.create_container(*args, **kwargs)


	def evaluate(self, source, **kwargs):
		if not self.is_ready: # no auto prepare
			raise self.NotReady
		info = self.create_eval_container(source=source, **kwargs)
		return self._evaluate(info)


	def _evaluate(self, info):
		raise NotImplementedError



class Fitable(AbstractFitable, AbstractEvaluatable):
	def create_fit_container(self, *args, **kwargs):
		return self.create_container(*args, **kwargs)


	def fit(self, source, **kwargs):
		self.prepare(source)
		info = self.create_fit_container(source=source, **kwargs)
		return self._fit(info)


	def _fit(self, info):
		raise NotImplementedError



class SimpleModel(Fitable, Evaluatable, Structured, AbstractModel):
	pass













from .top import tool, ToolKit, Context, Scope, Selection


def test_tool():
	"""
	This function tests the functionality of the 'tool' decorator and the 'gizmos' method.

	The 'tool' decorator is used to define two functions, 'f' and 'g', with 'a' and 'b' as their respective gizmos.
	The 'gizmos' method is then used to check if the gizmo of the function 'g' is correctly set to 'b'.
	"""

	@tool('a')  # The 'tool' decorator is used to define a function 'f' with 'a' as its gizmo.
	def f(x):
		"""
		This function takes an integer as input and returns the integer incremented by 1.

		Args:
			x (int): The input integer.

		Returns:
			int: The input integer incremented by 1.
		"""
		return x + 1

	assert f(1) == 2  # Asserts that the function 'f' correctly increments its input by 1.

	@tool('b')  # The 'tool' decorator is used to define a function 'g' with 'b' as its gizmo.
	def g(x, y, z):
		"""
		This function takes three integers as input and returns their sum.

		Args:
			x (int): The first input integer.
			y (int): The second input integer.
			z (int): The third input integer.

		Returns:
			int: The sum of the input integers.
		"""
		return x + y + z

	# Asserts that the gizmo of the function 'g' is correctly set to 'b'.
	assert list(g.gizmos()) == ['b'], f'gizmos: {list(g.gizmos())}'


def test_context():
	"""
	This function tests the functionality of the 'tool' decorator, the 'Context' class, and the 'include' method.

	The 'tool' decorator is used to define three functions, 'f', 'g', and 'f2', with 'y', 'z', and 'y' as their respective gizmos.
	The 'Context' class is used to create a context with the functions 'f' and 'g'.
	The 'include' method is then used to add the function 'f2' to the context.

	The function asserts that the context correctly maps 'x' to 'y' and that it updates correctly when the context's cache is cleared and 'f2' is included.
	"""

	@tool('y')  # The 'tool' decorator is used to define a function 'f' with 'y' as its gizmo.
	def f(x):
		"""
		This function takes an integer as input and returns the integer incremented by 1.

		Args:
			x (int): The input integer.

		Returns:
			int: The input integer incremented by 1.
		"""
		return x + 1

	@tool('z')  # The 'tool' decorator is used to define a function 'g' with 'z' as its gizmo.
	def g(x, y):
		"""
		This function takes two integers as input and returns their sum.

		Args:
			x (int): The first input integer.
			y (int): The second input integer.

		Returns:
			int: The sum of the input integers.
		"""
		return x + y

	@tool('y')  # The 'tool' decorator is used to define a function 'f2' with 'y' as its gizmo.
	def f2(y):
		"""
		This function takes an integer as input and returns the integer negated.

		Args:
			y (int): The input integer.

		Returns:
			int: The input integer negated.
		"""
		return -y

	ctx = Context(f, g)  # The 'Context' class is used to create a context with the functions 'f' and 'g'.

	ctx['x'] = 1  # The context maps 'x' to 'y'.
	assert ctx['y'] == 2  # Asserts that the context correctly maps 'x' to 'y'.

	ctx.clear_cache()  # The context's cache is cleared.
	ctx.include(f2)  # The function 'f2' is added to the context.

	ctx['x'] = 1  # The context maps 'x' to 'y'.
	assert ctx['y'] == -2  # Asserts that the context correctly maps 'x' to 'y' after 'f2' is included.


def _future_test_gizmo_dashes(): # this test is for the dash-gizmos
	"""
	This function tests the functionality of the 'tool' decorator, the 'gizmos' method, and the 'Context' class with gizmos that contain dashes.

	The 'tool' decorator is used to define a function 'f' with 'a-1' as its gizmo.
	The 'gizmos' method is then used to check if the gizmo of the function 'f' is correctly set to 'a_1'.
	The 'Context' class is used to create a context with the function 'f'.
	The function asserts that the context correctly maps 'a-1' to 1 and that it correctly identifies 'a-1' and 'a_1' as cached gizmos.
	"""

	@tool('a-1')  # The 'tool' decorator is used to define a function 'f' with 'a-1' as its gizmo.
	def f():
		"""
		This function returns the integer 1.

		Returns:
			int: The integer 1.
		"""
		return 1

	# Asserts that the gizmo of the function 'f' is correctly set to 'a_1'.
	assert list(f.gizmos()) == ['a_1'], f'gizmos: {list(f.gizmos())}'

	ctx = Context(f)  # The 'Context' class is used to create a context with the function 'f'.

	# Asserts that the context correctly maps 'a-1' to 1 and that it correctly identifies 'a-1' and 'a_1' as cached gizmos.
	assert ctx['a-1'] == 1, f'ctx["a-1"]: {ctx["a-1"]}'
	assert ctx.is_cached('a-1'), f'is_cached("a-1"): {ctx.is_cached("a-1")}'
	assert ctx.is_cached('a_1'), f'is_cached("a_1"): {ctx.is_cached("a_1")}'
	assert ctx['a_1'] == 1, f'ctx["a_1"]: {ctx["a_1"]}'


class _Kit1(ToolKit):
	"""
	The _Kit1 class is a subclass of ToolKit. It provides methods to handle gizmos 'y', 'z', and 'w'.
	"""

	@tool('y')  # The 'tool' decorator is used to define a function 'f' with 'y' as its gizmo.
	@staticmethod
	def f(x):
		"""
		This static method takes an integer as input and returns the integer incremented by 1.

		Args:
			x (int): The input integer.

		Returns:
			int: The input integer incremented by 1.
		"""
		return x + 1

	@tool('z')  # The 'tool' decorator is used to define a function 'g' with 'z' as its gizmo.
	def g(self, x, y):
		"""
		This method takes two integers as input and returns their sum.

		Args:
			x (int): The first input integer.
			y (int): The second input integer.

		Returns:
			int: The sum of the input integers.
		"""
		return x + y

	@tool('w')  # The 'tool' decorator is used to define a function 'h' with 'w' as its gizmo.
	@classmethod
	def h(cls, z):
		"""
		This class method takes an integer as input and returns the integer incremented by 2.

		Args:
			z (int): The input integer.

		Returns:
			int: The input integer incremented by 2.
		"""
		return z + 2

def test_crafty_kit():
	"""
	This function tests the functionality of the '_Kit1' class, the 'Context' class, and the 'clear_cache' method.

	The '_Kit1' class is instantiated and its methods 'f', 'g', and 'h' are tested.
	The 'Context' class is used to create a context with the '_Kit1' instance.
	The 'clear_cache' method is then used to clear the context's cache.

	The function asserts that the '_Kit1' instance correctly increments its input by 1 for method 'f', correctly sums its inputs for method 'g', and correctly increments its input by 2 for method 'h'.
	It also asserts that the context correctly maps 'x' to 'y', 'z', and 'w', and that it updates correctly when the context's cache is cleared.
	"""

	# Asserts that the 'f' method of the '_Kit1' class correctly increments its input by 1.
	assert _Kit1.f(1) == 2
	# Asserts that the 'h' method of the '_Kit1' class correctly increments its input by 2.
	assert _Kit1.h(1) == 3

	# The '_Kit1' class is instantiated.
	kit = _Kit1()
	# Asserts that the 'f' method of the '_Kit1' instance correctly increments its input by 1.
	assert kit.f(1) == 2
	# Asserts that the 'g' method of the '_Kit1' instance correctly sums its inputs.
	assert kit.g(1, 2) == 3
	# Asserts that the 'h' method of the '_Kit1' instance correctly increments its input by 2.
	assert kit.h(1) == 3

	# The 'Context' class is used to create a context with the '_Kit1' instance.
	ctx = Context(kit)

	# Asserts that the context correctly identifies the gizmos of the '_Kit1' instance.
	assert list(ctx.gizmos()) == ['y', 'z', 'w']

	# The context maps 'x' to 'y'.
	ctx['x'] = 1
	# Asserts that the context correctly maps 'x' to 'y'.
	assert ctx['y'] == 2
	# The context maps 'y' to 3.
	ctx['y'] = 3
	# Asserts that the context correctly maps 'y' to 3.
	assert ctx['y'] == 3
	# Asserts that the context correctly maps 'z' to 4.
	assert ctx['z'] == 4
	# Asserts that the context correctly maps 'w' to 6.
	assert ctx['w'] == 6

	# The context's cache is cleared.
	ctx.clear_cache()
	# The context maps 'x' to 10.
	ctx['x'] = 10
	# Asserts that the context correctly maps 'z' to 21 after the cache is cleared.
	assert ctx['z'] == 21
	# Asserts that the context correctly maps 'w' to 23 after the cache is cleared.
	assert ctx['w'] == 23

class _Kit2(_Kit1):  # Inherits all tools from the parent class by default
	"""
	The _Kit2 class is a subclass of _Kit1. It provides methods to handle gizmos 'y', 'z', and 'x'.
	It also includes a method to check the results of other methods.

	Attributes:
		_sign (int): A multiplier used in the get_x method. Defaults to 1.
	"""

	def __init__(self, sign=1):
		"""
		Initializes a new instance of the _Kit2 class.

		Args:
			sign (int): A multiplier used in the get_x method. Defaults to 1.
		"""
		super().__init__()
		self._sign = sign

	@tool('y')  # The 'tool' decorator is used to define a function 'change_y' with 'y' as its gizmo.
	def change_y(self, y):  # "Refinement" - chaining the tool implicitly
		"""
		This method takes an integer as input and returns the integer incremented by 10.

		Args:
			y (int): The input integer.

		Returns:
			int: The input integer incremented by 10.
		"""
		return y + 10

	@tool('x')  # The 'tool' decorator is used to define a function 'get_x' with 'x' as its gizmo.
	def get_x(self):
		"""
		This method returns the product of 100 and the _sign attribute.

		Returns:
			int: The product of 100 and the _sign attribute.
		"""
		return 100 * self._sign  # Freely use object attributes

	def check(self):  # Freely calling tools as methods
		"""
		This method returns the sum of the results of the 'f' method called with 9, the 'h' method called with 8, and the 'f' method called with 19.

		Returns:
			int: The sum of the results of the 'f' method called with 9, the 'h' method called with 8, and the 'f' method called with 19.
		"""
		return self.f(9) + type(self).h(8) + type(self).f(19)  # 40

	@tool('z')  # The 'tool' decorator is used to define a function 'g' with 'z' as its gizmo.
	def g(self, x):  # Overriding a tool (this will be registered, rather than the super method)
		"""
		This method takes an integer as input and returns the sum of the integer and itself.

		Args:
			x (int): The input integer.

		Returns:
			int: The sum of the input integer and itself.
		"""
		# Use with caution - it's recommended to use clear naming for the function
		return super().g(x, x)  # Super method can be called as usual

def test_crafty_kit_inheritance():
	"""
	This function tests the functionality of the '_Kit2' class, the 'Context' class, and the 'clear_cache' method.

	The '_Kit2' class is instantiated and its methods 'f', 'g', 'h', 'check', 'get_x', and 'change_y' are tested.
	The 'Context' class is used to create a context with the '_Kit2' instance.
	The 'clear_cache' method is then used to clear the context's cache.

	The function asserts that the '_Kit2' instance correctly increments its input by 1 for method 'f', correctly sums its input for method 'g', correctly increments its input by 2 for method 'h', correctly checks the results of other methods for method 'check', correctly gets the product of 100 and the _sign attribute for method 'get_x', and correctly increments its input by 10 for method 'change_y'.
	It also asserts that the context correctly maps 'x' to 'y', 'z', 'w', and 'x', and that it updates correctly when the context's cache is cleared and a new tool is included.
	"""

	# Asserts that the 'f' method of the '_Kit2' class correctly increments its input by 1.
	assert _Kit2.f(1) == 2
	# Asserts that the 'h' method of the '_Kit2' class correctly increments its input by 2.
	assert _Kit2.h(1) == 3

	# The '_Kit2' class is instantiated.
	kit = _Kit2()
	# Asserts that the 'f' method of the '_Kit2' instance correctly increments its input by 1.
	assert kit.f(1) == 2
	# Asserts that the 'g' method of the '_Kit2' instance correctly sums its input.
	assert kit.g(2) == 4
	# Asserts that the 'h' method of the '_Kit2' instance correctly increments its input by 2.
	assert kit.h(1) == 3
	# Asserts that the 'check' method of the '_Kit2' instance correctly checks the results of other methods.
	assert kit.check() == 40
	# Asserts that the 'get_x' method of the '_Kit2' instance correctly gets the product of 100 and the _sign attribute.
	assert kit.get_x() == 100
	# Asserts that the 'change_y' method of the '_Kit2' instance correctly increments its input by 10.
	assert kit.change_y(1) == 11

	# The 'Context' class is used to create a context with the '_Kit2' instance.
	ctx = Context(kit)

	# Asserts that the context correctly identifies the gizmos of the '_Kit2' instance.
	assert list(ctx.gizmos()) == ['y', 'z', 'w', 'x']

	# The context maps 'x' to 'y'.
	ctx['x'] = 100
	# Asserts that the context correctly maps 'x' to 'y'.
	assert ctx['y'] == 111
	# Asserts that the context correctly maps 'z' to 'w'.
	assert ctx['z'] == 200
	# Asserts that the context correctly maps 'w' to 'x'.
	assert ctx['w'] == 202

	# The context's cache is cleared.
	ctx.clear_cache()

	# A new tool is defined and included in the context.
	new_z = tool('z')(lambda: 1000)
	ctx.include(new_z)

	# Asserts that 'x' is not in the context's cache.
	assert 'x' not in ctx.cached()
	# Asserts that the context correctly maps 'y' to 'a'.
	assert ctx['y'] == 111
	# Asserts that 'x' is in the context's cache.
	assert 'x' in ctx.cached()
	# Asserts that the context correctly maps 'x' to 'y'.
	assert ctx['x'] == 100

	# Asserts that the context correctly maps 'z' to 'w' after the new tool is included.
	assert ctx['z'] == 1000
	# Asserts that the context correctly maps 'w' to 'x' after the new tool is included.
	assert ctx['w'] == 1002


class _Kit3(ToolKit):
	"""
	The _Kit3 class is a subclass of ToolKit. It provides methods to handle gizmos 'a', 'b', 'c', and 'd'.
	"""

	@tool('b')  # The 'tool' decorator is used to define a function 'f' with 'b' as its gizmo.
	@tool('a')  # The 'tool' decorator is used to define a function 'f' with 'a' as its gizmo.
	def f(self):
		"""
		This method returns the integer 1.

		Returns:
			int: The integer 1.
		"""
		return 1

	@tool('c')  # The 'tool' decorator is used to define a function 'g' with 'c' as its gizmo.
	@tool('b')  # The 'tool' decorator is used to define a function 'g' with 'b' as its gizmo.
	def g(self):
		"""
		This method returns the integer 2.

		Returns:
			int: The integer 2.
		"""
		return 2

	@tool('d')  # The 'tool' decorator is used to define a function 'h' with 'd' as its gizmo.
	@tool('c')  # The 'tool' decorator is used to define a function 'h' with 'c' as its gizmo.
	def h(self, b):
		"""
		This method takes an integer as input and returns the integer incremented by 10.

		Args:
			b (int): The input integer.

		Returns:
			int: The input integer incremented by 10.
		"""
		return b + 10


def test_nested_tools():
	"""
	This function tests the functionality of the '_Kit3' class and the 'Context' class.

	The '_Kit3' class is instantiated and a context is created with the '_Kit3' instance as its tool kit.
	The function asserts that the gizmos of the '_Kit3' instance are correctly identified and that the context correctly maps 'x' to 'y', 'z', 'w', and 'x'.
	"""

	# The '_Kit3' class is instantiated.
	kit = _Kit3()

	# Asserts that the gizmos of the '_Kit3' instance are correctly identified.
	assert list(kit.gizmos()) == ['a', 'b', 'c', 'd']

	# The 'Context' class is used to create a context with the '_Kit3' instance.
	ctx = Context(kit)

	# Asserts that the context correctly maps 'd' to 11.
	assert ctx['d'] == 11
	# Asserts that the context correctly maps 'c' to 2.
	assert ctx['c'] == 2
	# Asserts that the context correctly maps 'b' to 1.
	assert ctx['b'] == 1
	# Asserts that the context correctly maps 'a' to 1.
	assert ctx['a'] == 1


def test_scope():
	"""
	This function tests the functionality of the 'Scope' class and the 'Context' class with a '_Kit1' instance.

	The '_Kit1' class is instantiated and a scope is created with the '_Kit1' instance as its tool kit and a gizmo mapping from 'y' to 'a'.
	A context is then created with the scope as its tool kit.
	The function asserts that the gizmos of the scope are correctly identified and that the context correctly maps 'x' to 'a'.
	A new context is then created with a new scope as its tool kit, which has a gizmo mapping from 'y' to 'a' and 'x' to 'b'.
	The function asserts that the gizmos of the new scope are correctly identified and that the new context correctly maps 'b' to 'a' and 'z' to 'c'.
	"""

	# The '_Kit1' class is instantiated.
	kit = _Kit1()

	# The 'Scope' class is used to create a scope with the '_Kit1' instance and a gizmo mapping from 'y' to 'a'.
	scope = Scope(kit, gap={'y': 'a'})

	# Asserts that the gizmos of the scope are correctly identified.
	assert list(scope.gizmos()) == ['a', 'z', 'w']

	# The 'Context' class is used to create a context with the scope.
	ctx = Context(scope)

	# Asserts that the gizmos of the context are correctly identified.
	assert list(ctx.gizmos()) == ['a', 'z', 'w']

	# The context maps 'x' to 'y'.
	ctx['x'] = 1
	# Asserts that the context correctly maps 'x' to 'a'.
	assert ctx['a'] == 2

	# The 'Context' class is used to create a new context with a new scope, which has a gizmo mapping from 'y' to 'a' and 'x' to 'b'.
	ctx = Context(Scope(kit, gap={'y': 'a', 'x': 'b'}))

	# Asserts that the gizmos of the new scope are correctly identified.
	assert list(ctx.gizmos()) == ['a', 'z', 'w']

	# The context maps 'b' to 'x'.
	ctx['b'] = 1
	# Asserts that the context correctly maps 'b' to 'a'.
	assert ctx['a'] == 2
	# Asserts that the context correctly maps 'z' to 'c'.
	assert ctx['z'] == 3


def test_selection():
	"""
	This function tests the functionality of the 'Selection' class, the 'Context' class, and the '_Kit1' instance.

	The '_Kit1' class is instantiated and a selection is created with the '_Kit1' instance as its tool kit and a gizmo mapping from 'y' to 'a'.
	A context is then created with the selection as its tool kit.
	The function asserts that the gizmos of the selection are correctly identified and that the context correctly maps 'x' to 'y'.
	"""

	# The '_Kit1' class is instantiated.
	kit = _Kit1()

	# The 'Selection' class is used to create a selection with the '_Kit1' instance and a gizmo mapping from 'y' to 'a'.
	scope = Selection(kit, gap=['y'])

	# Asserts that the gizmos of the selection are correctly identified.
	assert list(scope.gizmos()) == ['y']

	# The 'Context' class is used to create a context with the selection.
	ctx = Context(scope)

	# Asserts that the gizmos of the context are correctly identified.
	assert list(ctx.gizmos()) == ['y']

	# The context maps 'x' to 'y'.
	ctx['x'] = 1

	# Asserts that the context correctly maps 'x' to 'y'.
	assert list(ctx.gizmos()) == ['x', 'y']

	# Asserts that the context correctly maps 'y' to 2.
	assert ctx['y'] == 2


def test_group_cache():
	"""
	This function tests the functionality of the 'tool' decorator, the 'Context' class, the 'Scope' class, and the '_Kit1' instance.

	The 'tool' decorator is used to define two functions, 'f' and 'g', with 'a' and 'x' as their respective gizmos.
	The 'Context' class is used to create a context with the functions 'f' and 'g'.
	The 'Scope' class is used to create a scope with the functions 'f' and 'g' and a gizmo mapping from 'a' to 'b'.
	The 'clear_cache' method is then used to clear the context's cache.

	The function asserts that the context correctly maps 'x' to 'y' and 'z', and that it updates correctly when the context's cache is cleared.
	"""

	counter = 0

	@tool('a')  # The 'tool' decorator is used to define a function 'f' with 'a' as its gizmo.
	def f():
		"""
		This function increments a counter and returns 1.

		Returns:
			int: The integer 1.
		"""
		nonlocal counter
		counter += 1
		return 1

	ctx = Context(f)  # The 'Context' class is used to create a context with the function 'f'.

	assert ctx['a'] == 1  # Asserts that the context correctly maps 'a' to 1.
	assert counter == 1  # Asserts that the counter is correctly incremented.
	assert 'a' in ctx.data  # Asserts that 'a' is in the context's data.
	assert ctx['a'] == 1  # Asserts that the context correctly maps 'a' to 1.
	assert counter == 1  # Asserts that the counter is not incremented.

	ctx = Context(Scope(f, gap={'a': 'b'}))  # The 'Scope' class is used to create a scope with the function 'f' and a gizmo mapping from 'a' to 'b'.

	assert not ctx.grabable('a')  # Asserts that 'a' is not grabable from the context.
	assert ctx.grabable('b')  # Asserts that 'b' is grabable from the context.
	assert ctx['b'] == 1  # Asserts that the context correctly maps 'b' to 1.
	assert counter == 2  # Asserts that the counter is correctly incremented.
	assert 'b' in ctx.data  # Asserts that 'b' is in the context's data.
	assert ctx['b'] == 1  # Asserts that the context correctly maps 'b' to 1.
	assert counter == 2  # Asserts that the counter is not incremented.

	ctx.clear_cache()  # The context's cache is cleared.

	assert ctx['b'] == 1  # Asserts that the context correctly maps 'b' to 1.
	assert counter == 3  # Asserts that the counter is correctly incremented.

	@tool('x')  # The 'tool' decorator is used to define a function 'g' with 'x' as its gizmo.
	def g(a):
		"""
		This function takes an integer as input and returns the product of the integer and 10.

		Args:
			a (int): The input integer.

		Returns:
			int: The product of the input integer and 10.
		"""
		return 10 * a

	ctx = Context(Scope(f, g, gap={'a': 'b'}))  # The 'Scope' class is used to create a scope with the functions 'f' and 'g' and a gizmo mapping from 'a' to 'b'.

	assert list(ctx.gizmos()) == ['b', 'x']  # Asserts that the gizmos of the context are correctly identified.

	assert ctx['x'] == 10  # Asserts that the context correctly maps 'x' to 10.
	assert counter == 4  # Asserts that the counter is correctly incremented.
	assert 'x' in ctx.data  # Asserts that 'x' is in the context's data.
	assert 'b' not in ctx.data  # Asserts that 'b' is not in the context's data.
	assert 'a' not in ctx.data  # Asserts that 'a' is not in the context's data.

	assert ctx.is_cached('x')  # Asserts that 'x' is cached in the context.
	assert ctx.is_cached('b')  # Asserts that 'b' is cached in the context.

	assert ctx['b'] == 1  # Asserts that the context correctly maps 'b' to 1.
	assert counter == 4  # Asserts that the counter is not incremented.

	ctx.clear_cache()  # The context's cache is cleared.

	assert ctx['b'] == 1  # Asserts that the context correctly maps 'b' to 1.
	assert counter == 5  # Asserts that the counter is correctly incremented.

	ctx.clear_cache()  # The context's cache is cleared.

	ctx['b'] = 2  # The context maps 'b' to 2.
	assert ctx['x'] == 20  # Asserts that the context correctly maps 'x' to 20.
	assert counter == 5  # Asserts that the counter is not incremented.

	ctx = Context(Scope(f, g))  # The 'Scope' class is used to create a scope with the functions 'f' and 'g'.

	assert list(ctx.gizmos()) == ['a', 'x']  # Asserts that the gizmos of the context are correctly identified.

	assert ctx['x'] == 10  # Asserts that the context correctly maps 'x' to 10.
	assert counter == 6  # Asserts that the counter is correctly incremented.
	assert 'x' in ctx.data  # Asserts that 'x' is in the context's data.
	assert 'a' not in ctx.data  # Asserts that 'a' is not in the context's data.
	assert ctx['a'] == 1  # Asserts that the context correctly maps 'a' to 1.
	assert counter == 6  # Asserts that the counter is not incremented.

	ctx.clear_cache()  # The context's cache is cleared.

	assert ctx['a'] == 1  # Asserts that the context correctly maps 'a' to 1.
	assert counter == 7  # Asserts that the counter is correctly incremented.

	ctx.clear_cache()  # The context's cache is cleared.

	ctx['a'] = 2  # The context maps 'a' to 2.
	assert ctx['x'] == 20  # Asserts that the context correctly maps 'x' to 20.
	assert counter == 7  # Asserts that the counter is not incremented.










from .imports import *

from .tools import *
from .kits import *
from .behaviors import *
from .contexts import *
from .quirks import *
from .quirky import *
from .spawning import *
from .tools import ToolDecorator, AutoToolDecorator

# added by codespaces


class tool(AutoToolDecorator):
	from_context = ToolDecorator



class TestKit(LoopyKit, MutableKit):
	def __init__(self, *tools: AbstractGadget, **kwargs):
		super().__init__(**kwargs)
		self.include(*tools)



class TestCraftyKitBase(CraftyKit):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self._process_crafts()



class TestContext(Cached, Gig, TestKit):#, Kit, AbstractContext):
	def tools(self, gizmo: Optional[str] = None) -> Iterator[AbstractGadget]:
		if gizmo is None:
			yield from filter_duplicates(chain.from_iterable(map(reversed, self._tools_table.values())))
		else:
			if gizmo not in self._tools_table:
				raise self._MissingGizmoError(gizmo)
			yield from reversed(self._tools_table[gizmo])



def test_tool():
	@tool('a')
	def f(x):
		return x + 1

	assert f(1) == 2

	@tool('b')
	def g(x, y, z):
		return x + y + z



def test_kit():
	@tool('y')
	def f(x):
		return x + 1

	@tool('z')
	def g(x, y):
		return x + y

	@tool('y')
	def f2(y):
		return -y

	ctx = TestContext(f, g)

	ctx['x'] = 1
	assert ctx['y'] == 2

	ctx.clear_cache()
	ctx.include(f2)

	ctx['x'] = 1
	assert ctx['y'] == -2



class TestCraftyKit(MutableKit, TestCraftyKitBase):
	@tool('y')
	@staticmethod
	def f(x):
		return x + 1

	@tool('z')
	def g(self, x, y):
		return x + y

	@tool('w')
	@classmethod
	def h(cls, z):
		return z + 2



def test_crafty_kit():
	assert TestCraftyKit.f(1) == 2
	assert TestCraftyKit.h(1) == 3

	kit = TestCraftyKit()
	assert kit.f(1) == 2
	assert kit.g(1, 2) == 3
	assert kit.h(1) == 3

	ctx = TestContext(kit)

	assert list(ctx.gizmos()) == ['y', 'z', 'w']

	ctx['x'] = 1
	assert ctx['y'] == 2
	ctx['y'] = 3
	assert ctx['y'] == 3
	assert ctx['z'] == 4
	assert ctx['w'] == 6

	ctx.clear_cache()
	ctx['x'] = 10
	assert ctx['z'] == 21
	assert ctx['w'] == 23



class TestCraftyKit2(TestCraftyKit): # by default inherits all tools from the parents
	def __init__(self, sign=1):
		super().__init__()
		self._sign = sign

	@tool('y') # tool replaced
	def change_y(self, y): # "refinement" - chaining the tool implicitly
		return y + 10

	@tool('x') # new tool added
	def get_x(self):
		return 100 * self._sign # freely use object attributes

	def check(self): # freely calling tools as methods
		return self.f(9) + type(self).h(8) + type(self).f(19) # 40

	def g(self, x): # overriding a tool (this will be registered, rather than the super method)
		# use with caution - it's recommended to use clear naming for the function
		return super().g(x, x) # super method can be called as usual



def test_crafty_kit_inheritance():

	assert TestCraftyKit2.f(1) == 2
	assert TestCraftyKit2.h(1) == 3

	kit = TestCraftyKit2()
	assert kit.f(1) == 2
	assert kit.g(2) == 4
	assert kit.h(1) == 3
	assert kit.check() == 40
	assert kit.get_x() == 100
	assert kit.change_y(1) == 11

	ctx = TestContext(kit)

	assert list(ctx.gizmos()) == ['y', 'z', 'w', 'x']

	assert ctx['x'] == 100
	assert ctx['y'] == 111
	assert ctx['z'] == 200
	assert ctx['w'] == 202

	ctx.clear_cache()

	@tool('z')
	def new_z():
		return 1000

	ctx.include(new_z)

	assert 'x' not in ctx.cached()
	assert ctx['y'] == 111
	assert 'x' in ctx.cached()
	assert ctx['x'] == 100

	assert ctx['z'] == 1000
	assert ctx['w'] == 1002



class LambdaTool(AbstractGadget):
	def __init__(self, fn, inp='input', out='output', **kwargs):
		if isinstance(inp, str):
			inp = [inp]
		super().__init__(**kwargs)
		self.fn = fn
		self.input_keys = inp
		self.output_key = out


	def __call__(self, *args, **kwargs):
		return self.fn(*args, **kwargs)


	def gizmos(self) -> Iterator[str]:
		yield self.output_key


	def grab_from(self, ctx: Optional[AbstractGig], gizmo: str) -> Any:
		if gizmo != self.output_key:
			raise GadgetFailedError(gizmo)
		# inputs = [ctx[g] for g in self.input_keys]
		inputs = [ctx.grab_from(ctx, g) for g in self.input_keys]
		return self.fn(*inputs)



class TestData(TestCraftyKitBase):
	@tool('data')
	def get_data(self):
		return 1, 1

	@tool('x')
	def get_x(self, data):
		return data[0]

	@tool('y_true')
	def get_y_true(self, data):
		return data[1]



class trait(AppliedTrait, SimpleQuirk):
	pass



class SuperModule(TestCraftyKitBase, Capable):
	augmentation = trait(apply={'input': 'x', 'output': 'x'})
	model = trait(apply={'input': 'x', 'output': 'y'})


	@tool('loss')
	@staticmethod
	def loss(y, y_true):
		return (y - y_true) ** 2


# class SuperModule2(SuperModule):
# 	augmentation = trait(apply={'input': 'original', 'output': 'augmented'})
# 	interpolator = trait(apply={'input1': 'original', 'input2': 'augmented', 'output': 'x'})
# 	model = trait(apply={'input': 'augmented', 'output': 'y'})



def test_trait():

	@tool('output')
	def augmentation(input):
		return input + 1

	@tool('output')
	def model(input):
		return input * 2

	sup = SuperModule(augmentation=augmentation, model=model)

	assert list(sup.gizmos()) == ['x', 'y', 'loss']

	ctx = TestContext(sup) # uses loopy getting

	assert list(ctx.gizmos()) == ['x', 'y', 'loss']

	ctx = TestContext(sup, TestData())

	assert ctx['loss'] == 9
	assert ctx['y'] == 4 # not 3, because the augmentation is applied first



class TestCrawler(SimpleCrawler, LoopyKit, MutableKit):
	def spawn(self, gizmo: Optional[str] = None) -> Iterator[Any]:
		if gizmo is not None:
			self.current.grab(gizmo)
		yield self.current
		yield from self

	def spawn_gizmo(self, gizmo: str) -> Iterator[str]:
		yield self[gizmo]
		for element in self:
			yield element[gizmo]



class TestDecision(AbstractDecision):
	def __init__(self, gizmo: str, choices: Iterable[Any] = (), **kwargs):
		if not isinstance(choices, (list, tuple)):
			choices = list(choices)
		super().__init__(**kwargs)
		self._choices = choices
		self._gizmo = gizmo

	def grab_from(self, ctx: Optional['AbstractGig'], gizmo: str) -> Any:
		if isinstance(ctx, AbstractCrawler):
			return ctx.select(self, gizmo)
		return super().grab_from(ctx, gizmo)

	def gizmos(self) -> Iterator[str]:
		yield self._gizmo

	def __len__(self):
		return len(self._choices)

	def choices(self, gizmo: str = None):
		yield from self._choices


def test_decisions():

	dec = TestDecision('simple', choices=[1, 2, 3])

	gz = list(dec.gizmos())
	assert gz == ['simple']

	ctx = TestCrawler().include(dec)

	gz = list(ctx.gizmos())
	assert gz == ['simple']

	x = ctx['simple']
	assert x == 1

	second = next(ctx)
	assert second['simple'] == 2

	rest = list(ctx)
	assert len(rest) == 1
	assert rest[0]['simple'] == 3


def test_tooled_decision():

	dec = TestDecision('chain', choices=[1, 2, 3])

	@tool('chain')
	def negative(chain):
		return -chain

	ctx = TestCrawler().include(negative, dec)

	assert list(ctx.gizmos()) == ['chain']

	itr = ctx.spawn('chain')

	frames = list(itr)
	results = []
	for f in frames:
		results.append(f['chain'])
	assert results == [-1, -2, -3]



def test_multi_decision():
	@tool('c')
	def c(a, b):
		return a + b

	a = TestDecision('a', choices=[10, 20, 30])
	b = TestDecision('b', choices=[1, 2, 3])

	ctx = TestCrawler().include(a, b, c)

	assert list(ctx.gizmos()) == ['a', 'b', 'c']

	itr = ctx.spawn_gizmo('c')

	results = list(itr)
	assert results == [11, 12, 13, 21, 22, 23, 31, 32, 33]










import sys, os
import numpy as np
# import torch
# import cv2
from io import BytesIO
# from PIL import Image



_pytorch_types = {
	# 'float': torch.float,
	# 'int': torch.int,
	# 'byte': torch.uint8,
	# 'long': torch.long,
	# 'float16': torch.float16,
	# 'double': torch.float64,
}
def pytorch_type(t):
	return _pytorch_types.get(t,t)



def jpeg_to_str(path):
	with open(path, 'rb') as f:
		return f.read()



def str_to_jpeg(s, ret_PIL=False):
	img = Image.open(BytesIO(s))
	if ret_PIL:
		return img
	return np.array(img)



















from .imports import *

from .quirks import *


class AbstractQuirky:
	def fill_quirks(self, fn, args=None, kwargs=None, **finder_kwargs) -> Dict[str, Any]:
		raise NotImplementedError


	def _extract_quirks(self, kwargs):
		raise NotImplementedError


	# def _existing_quirks(self, *, hidden=False):
	# 	for key, param in self.named_quirks(hidden=hidden):
	# 		if param.is_missing(self):
	# 			yield key, getattr(self, key)


	def has_quirk(self, key):
		return isinstance(self.get_quirk(key, None), AbstractQuirk)


	def get_quirk(self, key, default=unspecified_argument):
		raise NotImplementedError


	@agnostic
	def quirks(self, *, hidden=False):
		for key, val in self.named_quirks(hidden=hidden):
			yield val


	@agnostic
	def named_quirks(self, *, hidden=False): # N-O
		raise NotImplementedError


	@agnostic
	def quirk_names(self, *, hidden=False):
		for key, val in self.named_quirks(hidden=hidden):
			yield key



class Quirky(AbstractQuirky):
	def __init__(self, *args, **kwargs):
		params, remaining = self._extract_quirks(kwargs)
		super().__init__(*args, **remaining)
		for name, value in params.items():
			setattr(self, name, value)


	class _find_missing_quirk:
		def __init__(self, base, **kwargs):
			super().__init__(**kwargs)
			self.base = base


		def __call__(self, name, default=inspect.Parameter.empty):
			value = getattr(self.base, name, default)
			if value is inspect.Parameter.empty:
				raise KeyError(name)
			return value


	def fill_quirks(self, fn, args=None, kwargs=None, **finder_kwargs) -> Dict[str, Any]:
		params = extract_function_signature(fn, args=args, kwargs=kwargs, allow_positional=False,
		                                    default_fn=self._find_missing_quirk(self), **finder_kwargs)

		return params


	def _extract_quirks(self, kwargs):
		params = {}
		for name, _ in self.named_quirks(hidden=True):
			if name in kwargs:
				params[name] = kwargs.pop(name)
				# setattr(self, name, kwargs.pop(name))
		return params, kwargs


	@agnostic
	def get_hparam(self, key, default: Optional[Any] = unspecified_argument):
		val = inspect.getattr_static(self, key, unspecified_argument)
		if val is unspecified_argument:
			if default is unspecified_argument:
				raise AttributeError(f'{self} has no quirk {key!r}')
			return default
		return val


	@agnostic
	def named_quirks(self, *, hidden=False): # N-O
		owner = self if isinstance(self, type) else type(self)
		items = list(owner.__dict__.items())
		for key, val in reversed(items):
			if isinstance(val, AbstractQuirk) and (hidden or (isinstance(val, HiddenQuirk) and val.hidden(self))):
				yield key, val



class InheritableQuirky(AbstractQuirky):
	_my_inherited_quirks = None
	def __init_subclass__(cls, skip_inheritable_quirks=False, **kwargs):
		super().__init_subclass__(**kwargs)
		cls._my_inherited_quirks = []
		if not skip_inheritable_quirks:
			todo = [name for base in cls.__bases__ if issubclass(base, Quirky)
			        for name, param in base.named_quirks(hidden=True)
			        if isinstance(param, InheritableQuirk) and param.inherit(cls)]
			cls.inherit_quirks(*todo)


	@agnostic
	def named_quirks(self, *, hidden=False):
		yield from self._inherited_named_quirks(hidden=hidden)
		yield from super().named_quirks(hidden=hidden)


	@agnostic
	def inherit_quirks(self, *names: str) -> Union['InheritableQuirky', Type['InheritableQuirky']]: # TODO: return Self
		existing = set(self.quirk_names(hidden=True))
		self._my_inherited_quirks.extend(name for name in names if name not in existing)
		return self


	@agnostic
	def _inherited_named_quirks(self, *, hidden=False):
		if self._my_inherited_quirks is not None:
			for key in self._my_inherited_quirks:
				val = inspect.getattr_static(self, key, None)
				if isinstance(val, AbstractQuirk) and (hidden or (isinstance(val, HiddenQuirk) and val.hidden(self))):
					yield key, val



class Capable(InheritableQuirky, Quirky):
	pass






















class MissingValueError(AttributeError):
	pass



class MissingBuilderError(TypeError):
	pass



class NoProductFound(KeyError):
	pass



class MissingBranchError(NoProductFound):
	def __init__(self, branch, ident, msg=None):
		if msg is None:
			msg = f'Branch {branch!r} not found for ident {ident!r}'
		super().__init__(msg)



class InheritedHparamError(AttributeError):
	pass



class InvalidProductError(TypeError):
	pass




__version__ = "0.2.1"

from .core import *

# from . import util
# from .tools import *
# from .persistent import *
# from .features import *
# from .structure import *
# from .parameters import *
# from .modules import TrainableModel, SimpleModel, Trainer

# from .models import *
# from .data import *











from omniply.core.abstract import *



class AbstractSpaced(AbstractGadget):
	def space_from(self, ctx: 'AbstractGig', gizmo: str) -> Any:
		raise NotImplementedError


	def space_of(self, gizmo: str) -> Any:
		raise NotImplementedError


	def space_from(self, ctx: 'AbstractGig', gizmo: str) -> Any:
		raise NotImplementedError



class AbstractChangableSpaced(AbstractSpaced):
	def change_space_of(self, gizmo: str, space: Any) -> Any:
		raise NotImplementedError










from typing import Iterator, Optional, Any, Union
from omnibelt import unspecified_argument



class AbstractGizmo:
	"""
	AbstractGizmo is an abstract base class for custom labels of gizmos.

	Gizmos are the basic unit of data in `omni-ply`. They are used to represent both the input and output
	of gadgets. While the data itself may have any type, the label is by default a string, however if you would like
	to use a custom type of label, you can subclass AbstractGizmo and use that instead.
	"""

	def __new__(cls, label: Union[str, 'AbstractGizmo']):
		"""
		Constructor method for the AbstractGizmo class. It checks if the label is an instance of
		AbstractGizmo. If it is, it returns the label itself. Otherwise, it creates a new instance of the class.

		Args:
			label (Union[str, 'AbstractGizmo']): A string or an instance of AbstractGizmo.

		Returns:
			An instance of AbstractGizmo.
		"""
		if isinstance(label, AbstractGizmo):
			return label
		return super().__new__(cls)

	def __eq__(self, other) -> bool:
		"""
		Checks if the current instance is equal to another object. It does this by comparing
		their string representations.

		Args:
			other: The object to compare with.

		Returns:
			bool: True if the objects are equal, False otherwise.
		"""
		return str(self) == str(other)

	def __hash__(self) -> int:
		"""
		Returns the hash of the string representation of the current instance.

		Returns:
			int: The hash of the string representation of the current instance.
		"""
		return hash(str(self))


class AbstractGadgetError(Exception):
	"""
	AbstractGadgetError is an abstract base class for all exceptions that are raised by gadgets that should be
	handled by the inference engine.

	This class must be subclassed to create more specific exceptions raised by gadgets (see `errors.py`).
	"""

	@property
	def description(self) -> str:
		"""
		Returns a string description of the error. This method should be overridden by subclasses to provide the actual implementation.

		Returns:
			str: A string description of the error.
		"""
		raise NotImplementedError


class AbstractGadget:
	"""
	Gadgets define transformations of gizmos in `omni-ply`, and are the primary workhorses of the framework.
	Pretty much any class or function that you would like to use with the `omni-ply` structure, will be in the form
	of a subclass of this.

	AbstractGadget is an abstract base class for any custom gadget types.
	"""

	def gizmos(self) -> Iterator[str]:
		"""
		Generates all known products of this gadget.

		This method must be overridden by subclasses to provide the actual implementation.

		Returns:
			Iterator[str]: An iterator over the known products of this tool.
		"""
		raise NotImplementedError

	def grab_from(self, ctx: 'AbstractGig', gizmo: str) -> Any:
		"""
		Transforms the given context `ctx` to produce the specified `gizmo`, or raises ToolFailedError.
		The context can be expected to contain all the necessary input gizmos.

		This method should be overridden by subclasses to provide the actual implementation.

		Args:
			ctx (Optional['AbstractGig']): The context from which to grab any necessary input gizmos.
			gizmo (str): The gizmo that must be produced.

		Returns:
			Any: The specified output gizmo.

		Raises:
			ToolFailedError: If the gizmo cannot be grabbed (possibly because a necessary input gizmo is missing).
		"""
		raise NotImplementedError

	def grabable(self, gizmo: str) -> bool:
		"""
		Checks if this tool can produce the given gizmo.

		Args:
			gizmo (str): The gizmo to check.

		Returns:
			bool: True if this tool can produce the given gizmo, False otherwise.
		"""
		return gizmo in self.gizmos()

	def __repr__(self) -> str:
		"""
		Returns a string representation of this gadget.

		Returns:
			str: A string representation of this gadget.
		"""
		return f'{self.__class__.__name__}({", ".join(map(str, self.gizmos()))})'


class AbstractGaggle(AbstractGadget):
	"""
	A gaggle is a collection of gadgets, which because similar to a gadget, except that it can access sub-gadgets
	to produce specific gizmos.

	It provides methods to list all known gadgets and to return all known gadgets that can produce a given gizmo.
	This class is typically subclassed to create a custom types of gaggles.
	"""

	def gadgets(self, gizmo: Optional[str] = None) -> Iterator[AbstractGadget]:
		"""
		Lists all known gadgets under this gaggle in order of precedence.. If a gizmo is specified, it iterates over
		the gadgets that can produce the given gizmo. Note that, this will recursively iterate through all sub-gaggles
		and only yield the gadgets (ie. leaves of the tree).

		Args:
			gizmo (Optional[str]): If specified, yields only the gadgets that can produce this gizmo.

		Returns:
			Iterator[AbstractGadget]: An iterator over the known gadgets in this gaggle that can produce the
			specified gizmo.
		"""
		for gadget in self.vendors(gizmo):
			if isinstance(gadget, AbstractGaggle):
				yield from gadget.gadgets(gizmo)
			else:
				yield gadget

	def vendors(self, gizmo: Optional[str] = None) -> Iterator[AbstractGadget]:
		"""
		Lists all known sub-gadgets and sub-gaggles in this gaggle in order of precedence.

		Unlike `gadgets()`, this does not recursively iterate through sub-gaggles. It only considers the gadgets
		that are directly contained in this gaggle.

		This method must be overridden by subclasses to provide the actual implementation.

		Args:
			gizmo (Optional[str]): If specified, yields only the gadgets that can produce this gizmo.

		Returns:
			Iterator[AbstractGadget]: An iterator over the known gadgets that can directly produce the given gizmo.
		"""
		raise NotImplementedError


class AbstractMultiGadget(AbstractGaggle):
	"""
	AbstractMultiGadget is a special kind of gaggle that hides all sub-gadgets from being accessed through `gadgets()`
	and `vendors()`. Instead, it presents itself as a gadget that can produce all the products of the sub-gadgets.

	Generally, if you know before runtime what gizmos a gadget can produce, then it should just be a gadget, however,
	if you want to be able to dynamically add sub-gadgets, while still preventing delegation, then you can use this.

	"""

	def gadgets(self, gizmo: Optional[str] = None) -> Iterator[AbstractGadget]:
		"""
		Lists all known gadgets under this multi-gadget that can produce the given gizmo.
		Since this is a multi-gadget, it doesn't delegate to sub-gadgets, and instead yields itself.

		Args:
			gizmo (Optional[str]): If specified, yields only the gadgets that can produce this gizmo. In this case, it
			has no effect.

		Returns:
			Iterator[AbstractGadget]: An iterator over the known gadgets in this multi-gadget that can produce the
			specified gizmo. Since this is a multi-gadget, it yields only itself.
		"""
		yield self

	def vendors(self, gizmo: Optional[str] = None) -> Iterator[AbstractGadget]:
		"""
		Lists all known sub-gadgets and sub-gaggles in this multi-gadget that can produce the given gizmo.
		Since this is a multi-gadget, it doesn't delegate to sub-gadgets, and instead yields itself.

		Args:
			gizmo (Optional[str]): If specified, yields only the gadgets that can produce this gizmo. In this case, it
			checks if this multi-gadget can produce the gizmo.

		Returns:
			Iterator[AbstractGadget]: An iterator over the known gadgets that can directly produce the given gizmo. Since
			this is a multi-gadget, it yields itself.
		"""
		yield self


_unique_gig_default_value = object()


class AbstractGig(AbstractMultiGadget):
	"""
	Gigs are usually the top-level interface for users to use the inference engine of `omni-ply`. Gigs are a special
	kind of gaggle that takes ownership of a `grab_from()` call, rather than (usually silently) delegating to an
	appropriate gadget to produce the specified gizmo. This also means gigs are responsible providing the current
	context for gadgets (which often includes caching existing gizmos).
	"""

	def grab(self, gizmo: str, default: Any = _unique_gig_default_value):
		"""
		Convenience function for grab_from to match dict.get API. It returns the given gizmo from this gadget,
		or raises ToolFailedError if the gizmo cannot be grabbed and no default value is provided.

		Args:
			gizmo (str): The gizmo to grab.
			default (Any): The default value to return if the gizmo cannot be grabbed. If not specified,
			ToolFailedError is raised when the gizmo cannot be grabbed.

		Returns:
			Any: The grabbed gizmo, or the default value if the gizmo cannot be grabbed and a default value is provided.

		Raises:
			AbstractGadgetError: If the gizmo cannot be grabbed and no default value is provided.
		"""
		try:
			return self.grab_from(None, gizmo)
		except AbstractGadgetError:
			if default is _unique_gig_default_value:
				raise
			return default


class AbstractGang(AbstractGig):
	"""
	Gangs are a special kind of gig that relabels gizmos. It behaves a bit like a local/internal scope
	for its sub-gadgets, and can default to the global/external scope if necessary.

	This class must be typically subclassed to create a specific type of gang.
	"""

	def gizmo_from(self, gizmo: str) -> str:
		"""
		Converts external gizmo names to internal gizmo names used by sub-gadgets.

		This method must be overridden by subclasses to provide the actual implementation.

		Args:
			gizmo (str): The external gizmo name to convert.

		Returns:
			str: The internal gizmo name.
		"""
		raise NotImplementedError

	def gizmo_to(self, gizmo: str) -> str:
		"""
		Converts internal gizmo names used by sub-gadgets to external gizmo names.

		This method must be overridden by subclasses to provide the actual implementation.

		Args:
			gizmo (str): The internal gizmo name to convert.

		Returns:
			str: The external gizmo name.
		"""
		raise NotImplementedError



# Moguls -> Guru




from typing import Tuple, List, Dict, Optional, Union, Any, Callable, Sequence, Iterator, Iterable

from collections import OrderedDict
from omnibelt import get_printer, unspecified_argument, filter_duplicates

from .. import util
from ..structure import Metric, Sampler, Generator, spaces
from ..tools.kits import CraftyKit, DynamicKit
from ..tools.abstract import AbstractDynamicKit
from ..tools.context import DynamicContext

from .abstract import AbstractDataRouter, AbstractDataSource, AbstractBatchable, \
	AbstractCountableData, AbstractContext, AbstractTool
from .errors import UnknownSize


prt = get_printer(__file__)



class DataCollection(DynamicKit, AbstractBatchable, AbstractDataRouter):
	_BufferTable = OrderedDict

	def __init__(self, *, buffer_table=None, **kwargs):
		if buffer_table is None:
			buffer_table = self._BufferTable()
		super().__init__(**kwargs)
		self._buffers = buffer_table


	def include(self, *tools: AbstractTool, **buffers: AbstractDataSource) -> 'AbstractDynamicKit':
		for gizmo, buffer in buffers.items():
			self.register_buffer(gizmo, buffer)
		return super().include(*tools)


	def _validate_buffer(self, gizmo: str, buffer: AbstractTool):
		if not isinstance(buffer, AbstractDataSource):
			prt.warning(f'Expected buffer for {gizmo} in {self}, got: {buffer!r}')
		return buffer


	def buffer_names(self) -> Iterator[str]:
		yield from self._buffers.keys()


	def named_buffers(self) -> Iterator[Tuple[str, 'AbstractDataSource']]:
		for name in reversed(self._buffers):
			yield name, self.get_buffer(name)


	def get_buffer(self, name, default=unspecified_argument) -> 'AbstractDataSource':
		buffer = self._buffers.get(name, None)
		if name in self._buffers:
			if isinstance(buffer, str):
				return self.get_buffer(buffer, default=default)  # TODO: check for circular references
			return buffer
		if default is not unspecified_argument:
			return default
		raise self._MissingBuffer(name)


	# def space_of(self, gizmo: str):
	# 	buffer = self.get_buffer(gizmo, None)
	# 	if buffer is not None:
	# 		return buffer.space_of(gizmo)
	# 	return super().space_of(gizmo)


	# def _fingerprint_data(self):
	# 	# data = super()._fingerprint_data()
	# 	# if self.is_ready:
	# 	# 	data['buffers'] = {}
	# 	# 	for name, buffer in self.iter_named_buffers():
	# 	# 		data['buffers'][name] = buffer.fingerprint()
	# 	# return data
	# 	# return {'buffers': {name:buffer.fingerprint() for name, buffer in self.iter_named_buffers()}, 'ready': self.is_ready,
	# 	#         **super()._fingerprint_data()}
	# 	raise NotImplementedError

	def register_buffer(self, gizmo: str, buffer: Optional[AbstractTool] = None):
		self._buffers[gizmo] = self._validate_buffer(gizmo, buffer)
		return self


	def gizmos(self) -> Iterator[str]:
		yield from filter_duplicates(getattr(self, '_buffers', {}).keys(), super().gizmos())


	def vendors(self, gizmo: str):
		if gizmo in self._buffers:
			yield from self._buffers[gizmo].vendors(gizmo)
		yield from super().vendors(gizmo)


	def remove_buffer(self, gizmo: str):
		if gizmo in self._buffers:
			self._buffers.remove(gizmo)


	def rename_buffer(self, current: str, new: str):
		buffer = self.get_buffer(current, None)
		if buffer is not None:
			self.remove_buffer(current)
		self.register_buffer(new, buffer)



class CountableDataRouter(AbstractDataRouter, AbstractDataSource, AbstractCountableData):
	@property
	def size(self):
		return self._compute_size() if self.is_ready else self._expected_size()


	_UnknownSize = UnknownSize
	def _compute_size(self):
		for buffer in self.buffers():
			if isinstance(buffer, AbstractCountableData):
				try:
					return buffer.size
				except self._UnknownSize:
					pass
		return self._expected_size()


	def _expected_size(self):  # should be implemented by subclasses
		raise self._UnknownSize()



class BranchedDataRouter(DataCollection):
	def register_buffer(self, name, buffer=None, *, space=None, **kwargs):  # TODO: with delimiter for name
		raise NotImplementedError



class AutoCollection(DataCollection):
	_Buffer = None

	def _validate_buffer(self, gizmo: str, buffer: AbstractTool):
		if not isinstance(buffer, AbstractDataSource) and self._Buffer is not None:
			return self.register_buffer(gizmo, buffer)
		return super()._validate_buffer(gizmo, buffer)


	def register_buffer(self, gizmo, buffer: Optional[AbstractTool] = None, *,
	                    space: Optional[spaces.Dim] = None, **kwargs):
		if buffer is None:
			buffer = self._Buffer(space=space, **kwargs)
		elif not isinstance(buffer, AbstractDataSource):
			buffer = self._Buffer(buffer, space=space, **kwargs)
		return super().register_buffer(gizmo, buffer)



class PlaceholderCollection(DataCollection):
	def include(self, *tools: AbstractTool, **buffers: AbstractDataSource) -> 'AbstractDynamicKit':
		aliases = []
		remaining = []
		for tool in tools:
			(aliases if isinstance(tool, str) else remaining).append(tool)
		out = super().include(*remaining, **buffers)
		for alias in aliases:
			self.register_buffer(alias)
		return out



class AliasedCollection(DataCollection):
	def register_buffer_alias(self, name: str, *aliases: str):
		'''
		Registers aliases for a buffer.

		Args:
			name: original name of the buffer
			*aliases: all the new aliases

		Returns:

		'''
		for alias in aliases:
			self._buffers[alias] = name


	def has_gizmo(self, gizmo: str):
		alias = self._buffers[gizmo]
		return super().has_gizmo(gizmo) and (not isinstance(alias, str) or self.has_gizmo(alias))



########################################################################################################################



class Observation(AbstractDataRouter):  # SampleSource
	@property
	def din(self):
		return self.observation_space

	@property
	def observation_space(self):
		return self.space_of('observation')


class Supervised(Observation, Metric):
	@property
	def dout(self):
		return self.target_space

	@property
	def target_space(self):
		return self.space_of('target')

	def difference(self, a, b, standardize=None):
		return self.dout.difference(a, b, standardize=standardize)

	def measure(self, a, b, standardize=None):
		return self.dout.measure(a, b, standardize=standardize)

	def distance(self, a, b, standardize=None):
		return self.dout.distance(a, b, standardize=standardize)


class Labeled(Supervised):  # , alias={'target': 'label'}):
	@property
	def label_space(self):
		return self.space_of('label')


class Synthetic(Labeled):  # , alias={'label': 'mechanism'}): # TODO: include auto alias
	_distinct_mechanisms = True

	@property
	def mechanism_space(self):
		return self.space_of('mechanism')

	def transform_to_mechanisms(self, data):
		if not self._distinct_mechanisms:
			return data
		return self.mechanism_space.transform(data, self.label_space)

	def transform_to_labels(self, data):
		if not self._distinct_mechanisms:
			return data
		return self.label_space.transform(data, self.mechanism_space)












from typing import List, Dict, Tuple, Optional, Union, Any, Hashable, Sequence, Callable, Type, Iterable, Iterator
from omnibelt import unspecified_argument, get_printer
from omnibelt.crafts import AbstractCraft

from ..tools.abstract import AbstractScopeGenerator, AbstractScope, AbstractTool, ToolFailedError
from ..tools.assessments import Signatured, AbstractSignature, SimpleSignature
from ..tools.crafts import AbstractCrafty, ToolCraft, ReplaceableCraft
from ..tools.context import SimpleScope

from .abstract import AbstractSubmodule, AbstractModular, AbstractBuilder
from .errors import NoProductFound
from .hyperparameters import HyperparameterBase
from .building import get_builder, BuilderBase

prt = get_printer(__name__)



class SubmoduleBase(HyperparameterBase, AbstractSubmodule): # TODO: check builder for space (if none is provided)
	def __init__(self, default=unspecified_argument, *, typ=None, builder=None, **kwargs):
		super().__init__(default=default, **kwargs)
		# assert typ is None != builder is None, 'Either "typ" or "builder" must be provided'
		self.typ = typ
		self.builder = builder


	def copy(self, *, typ=None, builder=None, **kwargs):
		if typ is None:
			typ = self.typ
		if builder is None:
			builder = self.builder
		return super().copy(typ=typ, builder=builder, **kwargs)


	def init(self, instance):
		# TODO: (future) check for a special fget: spec -> builder for custom builders
		spec = instance.my_blueprint
		if not self.is_cached(instance):
			builder = self._create_default_builder(blueprint=None if spec is None else spec.for_builder())
			setattr(instance, self._default_builder_attr_code.format(self.attrname), builder)


	_default_builder_attr_code = '_temporary_default_builder_for_{}'
	def get_builder(self, owner):
		if isinstance(owner, type):
			return self._create_default_builder()
		if self.is_cached(owner):
			value = getattr(owner, self.attrname)
			if isinstance(value, AbstractBuilder):
				return value
		return getattr(owner, self._default_builder_attr_code.format(self.attrname), None)


	def validate(self, instance):
		value = getattr(instance, self.attrname, unspecified_argument)
		builder = getattr(instance, self._default_builder_attr_code.format(self.attrname), None)

		if builder is None:
			if value is unspecified_argument:
				raise self._MissingValueError(f'No product found for {self.attrname}')
		else:
			spec = instance.my_blueprint
			if spec is None:
				value = builder.build() if value is unspecified_argument else builder.validate(value)
			else:
				value = builder.build_with_spec(spec) if value is unspecified_argument \
					else builder.validate_with_spec(spec, value)
			setattr(instance, self.attrname, value)
			delattr(instance, self._default_builder_attr_code.format(self.attrname))


	_default_builder_type = None # TODO: simple builder that is given the product_base at __init__
	def _create_default_builder(self, *args, **kwargs) -> Optional[BuilderBase]:
		if self.builder is not None:
			builder = get_builder(self.builder) #if isinstance(self.builder, str) else self.builder
			return builder(*args, **kwargs)
		if self.typ is not None:
			return self._default_builder_type(self.typ)
		raise NotImplementedError('no info for default builder was provided') # TODO: make custom error type


	# def build_with(self, *args, **kwargs):
	# 	if self.typ is not None:
	# 		return self.typ(*args, **kwargs)
	# 	builder = self.get_builder()
	# 	if builder is None:
	# 		raise ValueError(f'No builder for {self}')
	# 	return builder.build(*args, **kwargs)
	#
	#
	# def build_with_spec(self, owner, spec):
	# 	builder = self.get_builder(blueprint=spec)
	# 	if builder is None:
	# 		raise ValueError(f'No builder for {self}')
	# 	product = builder.build()
	# 	spec.update_with(builder)
	# 	return product


	# def is_missing(self, owner):
	# 	return not (self.attrname in owner.__dict__ or self.fget is not None)


	# def missing_spaces(self, owner):
	# 	sub = getattr(owner, self.attrname, None)
	# 	if sub is not None:
	# 		yield from sub.missing_spaces()


	# def create_value(self, base, owner=None):  # TODO: maybe make thread-safe by using a lock
	# 	try:
	# 		return super().create_value(base, owner)
	# 	except self.MissingValueError:
	# 		builder = self.get_builder()
	# 		if builder is None:
	# 			raise
	# 		return builder.build()


##############

	# @agnostic
	# def full_spec(self, fmt='{}', fmt_rule='{parent}.{child}', include_machines=True):
	# 	for key, val in self.named_hyperparameters():
	# 		ident = fmt.format(key)
	# 		if isinstance(val, Machine):
	# 			builder = val.get_builder()
	# 			if include_machines or builder is None:
	# 				yield ident, val
	# 			if builder is not None:
	# 				if isinstance(builder, MachineParametrized):
	# 					yield from builder.full_spec(fmt=fmt_rule.format(parent=ident, child='{}'), fmt_rule=fmt_rule)
	# 				else:
	# 					for k, v in builder.plan():
	# 						yield fmt_rule.format(parent=ident, child=k), v
	# 		else:
	# 			yield ident, val



# class submodule(hparam):
# 	_registration_fn_name = 'register_submodule'


class SubmachineTool(ToolCraft, AbstractScopeGenerator):
	class Skill(ToolCraft.Skill):
		def space_of(self, gizmo: str):
			raise self._ToolFailedError(gizmo)


	def __init__(self, label: str, *, attrname=None, signature=None, **kwargs): # label is internal name (may be replaced)
		super().__init__(label, **kwargs)
		self.attrname = attrname
		self.signature = signature


	_Scope = SimpleScope
	def _create_scope(self, ctx, owner, attrname, sub):
		return self._Scope(sub, application=self.replacements)


	# def replace(self, replacements: Dict[str, str], **kwargs): # TODO: remove this
	# 	new = super().replace(replacements, **kwargs)
	# 	# if self.signature is not None:
	# 	# 	new.signature = self.signature.replace(replacements)
	# 	return new

	@property
	def wrapped(self):
		return None


	def _get_from(self, instance: AbstractCrafty, ctx, gizmo: str):
		sub = getattr(instance, self.attrname)

		if isinstance(sub, AbstractTool):
			scope = ctx.scope_for((instance, self.attrname), None)
			if scope is None:
				scope = self._create_scope(ctx, instance, self.attrname, sub)
				if scope is None:
					raise ValueError(f'No scope for {instance} and {self.attrname}')
				ctx.register_scope((instance, self.attrname), scope)

			return scope.get_from(ctx, gizmo)

		elif self.signature is not None:
			return self._automatic_get_from(sub, self.signature.replace(self.replacements), ctx)

		raise ToolFailedError(gizmo)


	def _automatic_get_from(self, sub, sig, ctx):
		if len(sig.inputs):
			vals = [ctx[inp] for inp in sig.inputs]
			return sub(*vals)
		elif 'size' in sig.meta:
			return sub(ctx.size)
		elif 'indices' in sig.meta:
			return sub(ctx.indices)
		elif 'index' in sig.meta:
			return sub(ctx.index)
		raise ToolFailedError(sig.output)


	def signatures(self, owner: Type[AbstractCrafty] = None):
		if self.signature is None:
			yield from super().signatures(owner)
		else:
			yield self.signature.replace(self.replacements, name=self.attrname)



class SubmachineBase(SubmoduleBase, Signatured, ReplaceableCraft):
	_Tool = SubmachineTool
	def __init__(self, default=unspecified_argument, *, application=None, replacements=None, **kwargs):
		if replacements is None:
			replacements = application
		if 'label' in kwargs:
			del kwargs['label'] # TODO: add warning about this
		# if 'replacements' in kwargs:
		# 	raise NotImplementedError('SubmachineBase does not support "replacements" (use "application" instead)')
		super().__init__(default=default, label=None, replacements=replacements, **kwargs)


	def copy(self, *, replacements=None, label=None, **kwargs):
		if replacements is None:
			replacements = self.replacements
		if label is None:
			label = self.label
		return super().copy(replacements=replacements, label=label, **kwargs)


	_default_builder_type = None # TODO: simple builder that is given the product_base at __init__
	def _create_default_builder(self, *args, application=None, **kwargs) -> Optional[BuilderBase]:
		if application is None:
			application = self.replacements
		return super()._create_default_builder(*args, application=application, **kwargs)


	def _expected_signatures(self, owner=None):
		if not isinstance(owner, type) and self.is_cached(owner):
			value = getattr(owner, self.attrname)
			if isinstance(value, Signatured):
				yield from value.signatures()
				return
		builder = self.get_builder(owner)
		if builder is not None:
			yield from builder.product_signatures()


	def emit_craft_items(self, owner=None):
		if self.replacements is not None:
			for signature in self._expected_signatures(owner=owner):
				if signature.output in self.replacements:
					yield self._Tool(signature.output, attrname=self.attrname, replacements=self.replacements,
					                 signature=signature)








